from contextlib import contextmanager

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    package_version,
)

try:
    from graphene.types import schema as graphene_schema  # type: ignore
except ImportError:
    raise DidNotEnable("graphene is not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any, Dict, Union

    from graphene.language.source import Source  # type: ignore
    from graphql.execution import ExecutionResult
    from graphql.type import GraphQLSchema

    from sentry_sdk._types import Event


class GrapheneIntegration(Integration):
    identifier = "graphene"

    @staticmethod
    def setup_once() -> None:
        version = package_version("graphene")
        _check_minimum_version(GrapheneIntegration, version)

        _patch_graphql()


def _patch_graphql() -> None:
    old_graphql_sync = graphene_schema.graphql_sync
    old_graphql_async = graphene_schema.graphql

    @ensure_integration_enabled(GrapheneIntegration, old_graphql_sync)
    def _sentry_patched_graphql_sync(
        schema: "GraphQLSchema",
        source: "Union[str, Source]",
        *args: "Any",
        **kwargs: "Any",
    ) -> "ExecutionResult":
        scope = sentry_sdk.get_isolation_scope()
        scope.add_event_processor(_event_processor)

        with graphql_span(schema, source, kwargs):
            result = old_graphql_sync(schema, source, *args, **kwargs)

        with capture_internal_exceptions():
            client = sentry_sdk.get_client()
            for error in result.errors or []:
                event, hint = event_from_exception(
                    error,
                    client_options=client.options,
                    mechanism={
                        "type": GrapheneIntegration.identifier,
                        "handled": False,
                    },
                )
                sentry_sdk.capture_event(event, hint=hint)

        return result

    async def _sentry_patched_graphql_async(
        schema: "GraphQLSchema",
        source: "Union[str, Source]",
        *args: "Any",
        **kwargs: "Any",
    ) -> "ExecutionResult":
        integration = sentry_sdk.get_client().get_integration(GrapheneIntegration)
        if integration is None:
            return await old_graphql_async(schema, source, *args, **kwargs)

        scope = sentry_sdk.get_isolation_scope()
        scope.add_event_processor(_event_processor)

        with graphql_span(schema, source, kwargs):
            result = await old_graphql_async(schema, source, *args, **kwargs)

        with capture_internal_exceptions():
            client = sentry_sdk.get_client()
            for error in result.errors or []:
                event, hint = event_from_exception(
                    error,
                    client_options=client.options,
                    mechanism={
                        "type": GrapheneIntegration.identifier,
                        "handled": False,
                    },
                )
                sentry_sdk.capture_event(event, hint=hint)

        return result

    graphene_schema.graphql_sync = _sentry_patched_graphql_sync
    graphene_schema.graphql = _sentry_patched_graphql_async


def _event_processor(event: "Event", hint: "Dict[str, Any]") -> "Event":
    if should_send_default_pii():
        request_info = event.setdefault("request", {})
        request_info["api_target"] = "graphql"

    elif event.get("request", {}).get("data"):
        del event["request"]["data"]

    return event


@contextmanager
def graphql_span(
    schema: "GraphQLSchema", source: "Union[str, Source]", kwargs: "Dict[str, Any]"
) -> "Generator[None, None, None]":
    operation_name = kwargs.get("operation_name") or "<unknown graphql operation>"

    operation_type = "query"
    op = OP.GRAPHQL_QUERY
    if source.strip().startswith("mutation"):
        operation_type = "mutation"
        op = OP.GRAPHQL_MUTATION
    elif source.strip().startswith("subscription"):
        operation_type = "subscription"
        op = OP.GRAPHQL_SUBSCRIPTION

    sentry_sdk.add_breadcrumb(
        crumb={
            "data": {
                "operation_name": operation_name,
                "operation_type": operation_type,
            },
            "category": "graphql.operation",
        },
    )

    is_span_streaming_enabled = has_span_streaming_enabled(
        sentry_sdk.get_client().options
    )

    if is_span_streaming_enabled:
        if sentry_sdk.traces.get_current_span() is None:
            yield
            return

        additional_attributes = {}
        if should_send_default_pii():
            additional_attributes["graphql.document"] = source

        _graphql_span = sentry_sdk.traces.start_span(
            name=operation_name,
            attributes={
                "sentry.op": op,
                "graphql.operation.name": operation_name,
                "graphql.operation.type": operation_type,
                **additional_attributes,
            },
        )
    else:
        _graphql_span = sentry_sdk.start_span(op=op, name=operation_name)

        if should_send_default_pii():
            _graphql_span.set_data("graphql.document", source)
        _graphql_span.set_data("graphql.operation.name", operation_name)
        _graphql_span.set_data("graphql.operation.type", operation_type)

        _graphql_span.__enter__()

    try:
        yield
    finally:
        if is_span_streaming_enabled:
            _graphql_span.end()  # type: ignore
        else:
            _graphql_span.__exit__(None, None, None)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.crons import MonitorStatus, capture_checkin
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.integrations.celery.utils import (
    _get_humanized_interval,
    _now_seconds_since_epoch,
)
from sentry_sdk.utils import (
    logger,
    match_regex_list,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional, TypeVar, Union

    from sentry_sdk._types import (
        MonitorConfig,
        MonitorConfigScheduleType,
        MonitorConfigScheduleUnit,
    )

    F = TypeVar("F", bound=Callable[..., Any])


try:
    from celery import Celery, Task  # type: ignore
    from celery.beat import Scheduler  # type: ignore
    from celery.schedules import crontab, schedule  # type: ignore
    from celery.signals import (  # type: ignore
        task_failure,
        task_retry,
        task_success,
    )
except ImportError:
    raise DidNotEnable("Celery not installed")

try:
    from redbeat.schedulers import RedBeatScheduler  # type: ignore
except ImportError:
    RedBeatScheduler = None


def _get_headers(task: "Task") -> "dict[str, Any]":
    headers = task.request.get("headers") or {}

    # flatten nested headers
    if "headers" in headers:
        headers.update(headers["headers"])
        del headers["headers"]

    headers.update(task.request.get("properties") or {})

    return headers


def _get_monitor_config(
    celery_schedule: "Any", app: "Celery", monitor_name: str
) -> "MonitorConfig":
    monitor_config: "MonitorConfig" = {}
    schedule_type: "Optional[MonitorConfigScheduleType]" = None
    schedule_value: "Optional[Union[str, int]]" = None
    schedule_unit: "Optional[MonitorConfigScheduleUnit]" = None

    if isinstance(celery_schedule, crontab):
        schedule_type = "crontab"
        schedule_value = (
            "{0._orig_minute} "
            "{0._orig_hour} "
            "{0._orig_day_of_month} "
            "{0._orig_month_of_year} "
            "{0._orig_day_of_week}".format(celery_schedule)
        )
    elif isinstance(celery_schedule, schedule):
        schedule_type = "interval"
        (schedule_value, schedule_unit) = _get_humanized_interval(
            celery_schedule.seconds
        )

        if schedule_unit == "second":
            logger.warning(
                "Intervals shorter than one minute are not supported by Sentry Crons. Monitor '%s' has an interval of %s seconds. Use the `exclude_beat_tasks` option in the celery integration to exclude it.",
                monitor_name,
                schedule_value,
            )
            return {}

    else:
        logger.warning(
            "Celery schedule type '%s' not supported by Sentry Crons.",
            type(celery_schedule),
        )
        return {}

    monitor_config["schedule"] = {}
    monitor_config["schedule"]["type"] = schedule_type
    monitor_config["schedule"]["value"] = schedule_value

    if schedule_unit is not None:
        monitor_config["schedule"]["unit"] = schedule_unit

    monitor_config["timezone"] = (
        (
            hasattr(celery_schedule, "tz")
            and celery_schedule.tz is not None
            and str(celery_schedule.tz)
        )
        or app.timezone
        or "UTC"
    )

    return monitor_config


def _apply_crons_data_to_schedule_entry(
    scheduler: "Any",
    schedule_entry: "Any",
    integration: "sentry_sdk.integrations.celery.CeleryIntegration",
) -> None:
    """
    Add Sentry Crons information to the schedule_entry headers.
    """
    if not integration.monitor_beat_tasks:
        return

    monitor_name = schedule_entry.name

    task_should_be_excluded = match_regex_list(
        monitor_name, integration.exclude_beat_tasks
    )
    if task_should_be_excluded:
        return

    celery_schedule = schedule_entry.schedule
    app = scheduler.app

    monitor_config = _get_monitor_config(celery_schedule, app, monitor_name)

    is_supported_schedule = bool(monitor_config)
    if not is_supported_schedule:
        return

    headers = schedule_entry.options.pop("headers", {})
    headers.update(
        {
            "sentry-monitor-slug": monitor_name,
            "sentry-monitor-config": monitor_config,
        }
    )

    check_in_id = capture_checkin(
        monitor_slug=monitor_name,
        monitor_config=monitor_config,
        status=MonitorStatus.IN_PROGRESS,
    )
    headers.update({"sentry-monitor-check-in-id": check_in_id})

    # Set the Sentry configuration in the options of the ScheduleEntry.
    # Those will be picked up in `apply_async` and added to the headers.
    schedule_entry.options["headers"] = headers


def _wrap_beat_scheduler(
    original_function: "Callable[..., Any]",
) -> "Callable[..., Any]":
    """
    Makes sure that:
    - a new Sentry trace is started for each task started by Celery Beat and
      it is propagated to the task.
    - the Sentry Crons information is set in the Celery Beat task's
      headers so that is monitored with Sentry Crons.

    After the patched function is called,
    Celery Beat will call apply_async to put the task in the queue.
    """
    # Patch only once
    # Can't use __name__ here, because some of our tests mock original_apply_entry
    already_patched = "sentry_patched_scheduler" in str(original_function)
    if already_patched:
        return original_function

    from sentry_sdk.integrations.celery import CeleryIntegration

    def sentry_patched_scheduler(*args: "Any", **kwargs: "Any") -> None:
        integration = sentry_sdk.get_client().get_integration(CeleryIntegration)
        if integration is None:
            return original_function(*args, **kwargs)

        # Tasks started by Celery Beat start a new Trace
        scope = sentry_sdk.get_isolation_scope()
        scope.set_new_propagation_context()
        scope._name = "celery-beat"

        scheduler, schedule_entry = args
        _apply_crons_data_to_schedule_entry(scheduler, schedule_entry, integration)

        return original_function(*args, **kwargs)

    return sentry_patched_scheduler


def _patch_beat_apply_entry() -> None:
    Scheduler.apply_entry = _wrap_beat_scheduler(Scheduler.apply_entry)


def _patch_redbeat_apply_async() -> None:
    if RedBeatScheduler is None:
        return

    RedBeatScheduler.apply_async = _wrap_beat_scheduler(RedBeatScheduler.apply_async)


def _setup_celery_beat_signals(monitor_beat_tasks: bool) -> None:
    if monitor_beat_tasks:
        task_success.connect(crons_task_success)
        task_failure.connect(crons_task_failure)
        task_retry.connect(crons_task_retry)


def crons_task_success(sender: "Task", **kwargs: "dict[Any, Any]") -> None:
    logger.debug("celery_task_success %s", sender)
    headers = _get_headers(sender)

    if "sentry-monitor-slug" not in headers:
        return

    monitor_config = headers.get("sentry-monitor-config", {})

    start_timestamp_s = headers.get("sentry-monitor-start-timestamp-s")

    capture_checkin(
        monitor_slug=headers["sentry-monitor-slug"],
        monitor_config=monitor_config,
        check_in_id=headers["sentry-monitor-check-in-id"],
        duration=(
            _now_seconds_since_epoch() - float(start_timestamp_s)
            if start_timestamp_s
            else None
        ),
        status=MonitorStatus.OK,
    )


def crons_task_failure(sender: "Task", **kwargs: "dict[Any, Any]") -> None:
    logger.debug("celery_task_failure %s", sender)
    headers = _get_headers(sender)

    if "sentry-monitor-slug" not in headers:
        return

    monitor_config = headers.get("sentry-monitor-config", {})

    start_timestamp_s = headers.get("sentry-monitor-start-timestamp-s")

    capture_checkin(
        monitor_slug=headers["sentry-monitor-slug"],
        monitor_config=monitor_config,
        check_in_id=headers["sentry-monitor-check-in-id"],
        duration=(
            _now_seconds_since_epoch() - float(start_timestamp_s)
            if start_timestamp_s
            else None
        ),
        status=MonitorStatus.ERROR,
    )


def crons_task_retry(sender: "Task", **kwargs: "dict[Any, Any]") -> None:
    logger.debug("celery_task_retry %s", sender)
    headers = _get_headers(sender)

    if "sentry-monitor-slug" not in headers:
        return

    monitor_config = headers.get("sentry-monitor-config", {})

    start_timestamp_s = headers.get("sentry-monitor-start-timestamp-s")

    capture_checkin(
        monitor_slug=headers["sentry-monitor-slug"],
        monitor_config=monitor_config,
        check_in_id=headers["sentry-monitor-check-in-id"],
        duration=(
            _now_seconds_since_epoch() - float(start_timestamp_s)
            if start_timestamp_s
            else None
        ),
        status=MonitorStatus.ERROR,
    )



import asyncio
import os
import threading
from abc import ABC, abstractmethod
from time import sleep, time
from typing import TYPE_CHECKING

from sentry_sdk._queue import FullError, Queue
from sentry_sdk.consts import DEFAULT_QUEUE_SIZE
from sentry_sdk.utils import logger, mark_sentry_task_internal

if TYPE_CHECKING:
    from typing import Any, Callable, Optional


_TERMINATOR = object()


class Worker(ABC):
    """Base class for all workers."""

    @property
    @abstractmethod
    def is_alive(self) -> bool:
        """Whether the worker is alive and running."""
        pass

    @abstractmethod
    def kill(self) -> None:
        """Kill the worker. It will not process any more events."""
        pass

    def flush(
        self, timeout: float, callback: "Optional[Callable[[int, float], Any]]" = None
    ) -> None:
        """Flush the worker, blocking until done or timeout is reached."""
        return None

    @abstractmethod
    def full(self) -> bool:
        """Whether the worker's queue is full."""
        pass

    @abstractmethod
    def submit(self, callback: "Callable[[], Any]") -> bool:
        """Schedule a callback. Returns True if queued, False if full."""
        pass


class BackgroundWorker(Worker):
    def __init__(self, queue_size: int = DEFAULT_QUEUE_SIZE) -> None:
        self._queue: "Queue" = Queue(queue_size)
        self._lock = threading.Lock()
        self._thread: "Optional[threading.Thread]" = None
        self._thread_for_pid: "Optional[int]" = None

    @property
    def is_alive(self) -> bool:
        if self._thread_for_pid != os.getpid():
            return False
        if not self._thread:
            return False
        return self._thread.is_alive()

    def _ensure_thread(self) -> None:
        if not self.is_alive:
            self.start()

    def _timed_queue_join(self, timeout: float) -> bool:
        deadline = time() + timeout
        queue = self._queue

        queue.all_tasks_done.acquire()

        try:
            while queue.unfinished_tasks:
                delay = deadline - time()
                if delay <= 0:
                    return False
                queue.all_tasks_done.wait(timeout=delay)

            return True
        finally:
            queue.all_tasks_done.release()

    def start(self) -> None:
        with self._lock:
            if not self.is_alive:
                self._thread = threading.Thread(
                    target=self._target, name="sentry-sdk.BackgroundWorker"
                )
                self._thread.daemon = True
                try:
                    self._thread.start()
                    self._thread_for_pid = os.getpid()
                except RuntimeError:
                    # At this point we can no longer start because the interpreter
                    # is already shutting down.  Sadly at this point we can no longer
                    # send out events.
                    self._thread = None

    def kill(self) -> None:
        """
        Kill worker thread. Returns immediately. Not useful for
        waiting on shutdown for events, use `flush` for that.
        """
        logger.debug("background worker got kill request")
        with self._lock:
            if self._thread:
                try:
                    self._queue.put_nowait(_TERMINATOR)
                except FullError:
                    logger.debug("background worker queue full, kill failed")

                self._thread = None
                self._thread_for_pid = None

    def flush(self, timeout: float, callback: "Optional[Any]" = None) -> None:
        logger.debug("background worker got flush request")
        with self._lock:
            if self.is_alive and timeout > 0.0:
                self._wait_flush(timeout, callback)
        logger.debug("background worker flushed")

    def full(self) -> bool:
        return self._queue.full()

    def _wait_flush(self, timeout: float, callback: "Optional[Any]") -> None:
        initial_timeout = min(0.1, timeout)
        if not self._timed_queue_join(initial_timeout):
            pending = self._queue.qsize() + 1
            logger.debug("%d event(s) pending on flush", pending)
            if callback is not None:
                callback(pending, timeout)

            if not self._timed_queue_join(timeout - initial_timeout):
                pending = self._queue.qsize() + 1
                logger.error("flush timed out, dropped %s events", pending)

    def submit(self, callback: "Callable[[], Any]") -> bool:
        self._ensure_thread()
        try:
            self._queue.put_nowait(callback)
            return True
        except FullError:
            return False

    def _target(self) -> None:
        while True:
            callback = self._queue.get()
            try:
                if callback is _TERMINATOR:
                    break
                try:
                    callback()
                except Exception:
                    logger.error("Failed processing job", exc_info=True)
            finally:
                self._queue.task_done()
            sleep(0)


class AsyncWorker(Worker):
    def __init__(self, queue_size: int = DEFAULT_QUEUE_SIZE) -> None:
        self._queue: "Optional[asyncio.Queue[Any]]" = None
        self._queue_size = queue_size
        self._task: "Optional[asyncio.Task[None]]" = None
        # Event loop needs to remain in the same process
        self._task_for_pid: "Optional[int]" = None
        self._loop: "Optional[asyncio.AbstractEventLoop]" = None
        # Track active callback tasks so they have a strong reference and can be cancelled on kill
        self._active_tasks: "set[asyncio.Task[None]]" = set()

    @property
    def is_alive(self) -> bool:
        if self._task_for_pid != os.getpid():
            return False
        if not self._task or not self._loop:
            return False
        return self._loop.is_running() and not self._task.done()

    def kill(self) -> None:
        if self._task:
            # Cancel the main consumer task to prevent duplicate consumers
            self._task.cancel()
            # Also cancel any active callback tasks
            # Avoid modifying the set while cancelling tasks
            tasks_to_cancel = set(self._active_tasks)
            for task in tasks_to_cancel:
                task.cancel()
            self._active_tasks.clear()
            self._loop = None
            self._task = None
            self._task_for_pid = None

    def start(self) -> None:
        if not self.is_alive:
            try:
                self._loop = asyncio.get_running_loop()
                # Always create a fresh queue on start to avoid stale items
                self._queue = asyncio.Queue(maxsize=self._queue_size)
                with mark_sentry_task_internal():
                    self._task = self._loop.create_task(self._target())
                self._task_for_pid = os.getpid()
            except RuntimeError:
                # There is no event loop running
                logger.warning("No event loop running, async worker not started")
                self._loop = None
                self._task = None
                self._task_for_pid = None

    def full(self) -> bool:
        if self._queue is None:
            return True
        return self._queue.full()

    def _ensure_task(self) -> None:
        if not self.is_alive:
            self.start()

    async def _wait_flush(
        self, timeout: float, callback: "Optional[Any]" = None
    ) -> None:
        if not self._loop or not self._loop.is_running() or self._queue is None:
            return

        initial_timeout = min(0.1, timeout)

        # Timeout on the join
        try:
            await asyncio.wait_for(self._queue.join(), timeout=initial_timeout)
        except asyncio.TimeoutError:
            pending = self._queue.qsize() + len(self._active_tasks)
            logger.debug("%d event(s) pending on flush", pending)
            if callback is not None:
                callback(pending, timeout)

            try:
                remaining_timeout = timeout - initial_timeout
                await asyncio.wait_for(self._queue.join(), timeout=remaining_timeout)
            except asyncio.TimeoutError:
                pending = self._queue.qsize() + len(self._active_tasks)
                logger.error("flush timed out, dropped %s events", pending)

    def flush(  # type: ignore[override]
        self, timeout: float, callback: "Optional[Any]" = None
    ) -> "Optional[asyncio.Task[None]]":
        if self.is_alive and timeout > 0.0 and self._loop and self._loop.is_running():
            with mark_sentry_task_internal():
                return self._loop.create_task(self._wait_flush(timeout, callback))
        return None

    def submit(self, callback: "Callable[[], Any]") -> bool:
        self._ensure_task()
        if self._queue is None:
            return False
        try:
            self._queue.put_nowait(callback)
            return True
        except asyncio.QueueFull:
            return False

    async def _target(self) -> None:
        if self._queue is None:
            return
        try:
            while True:
                callback = await self._queue.get()
                if callback is _TERMINATOR:
                    self._queue.task_done()
                    break
                # Firing tasks instead of awaiting them allows for concurrent requests
                with mark_sentry_task_internal():
                    task = asyncio.create_task(self._process_callback(callback))
                # Create a strong reference to the task so it can be cancelled on kill
                # and does not get garbage collected while running
                self._active_tasks.add(task)
                # Capture queue ref at dispatch time so done callbacks use the
                # correct queue even if kill()/start() replace self._queue.
                queue_ref = self._queue
                task.add_done_callback(lambda t: self._on_task_complete(t, queue_ref))
                # Yield to let the event loop run other tasks
                await asyncio.sleep(0)
        except asyncio.CancelledError:
            pass  # Expected during kill()

    async def _process_callback(self, callback: "Callable[[], Any]") -> None:
        # Callback is an async coroutine, need to await it
        await callback()

    def _on_task_complete(
        self,
        task: "asyncio.Task[None]",
        queue: "Optional[asyncio.Queue[Any]]" = None,
    ) -> None:
        try:
            task.result()
        except asyncio.CancelledError:
            pass  # Task was cancelled, expected during shutdown
        except Exception:
            logger.error("Failed processing job", exc_info=True)
        finally:
            # Mark the task as done and remove it from the active tasks set
            # Use the queue reference captured at dispatch time, not self._queue,
            # to avoid calling task_done() on a different queue after kill()/start().
            if queue is not None:
                queue.task_done()
            self._active_tasks.discard(task)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.integrations.grpc.consts import SPAN_ORIGIN
from sentry_sdk.traces import SegmentNameSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import event_from_exception

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Any, Optional


try:
    import grpc
    from grpc import HandlerCallDetails, RpcMethodHandler
    from grpc.aio import AbortError, ServicerContext
except ImportError:
    raise DidNotEnable("grpcio is not installed")


class ServerInterceptor(grpc.aio.ServerInterceptor):  # type: ignore
    def __init__(
        self: "ServerInterceptor",
        find_name: "Callable[[ServicerContext], str] | None" = None,
    ) -> None:
        self._custom_find_name = find_name

        super().__init__()

    async def intercept_service(
        self: "ServerInterceptor",
        continuation: "Callable[[HandlerCallDetails], Awaitable[RpcMethodHandler]]",
        handler_call_details: "HandlerCallDetails",
    ) -> "Optional[Awaitable[RpcMethodHandler]]":
        handler = await continuation(handler_call_details)
        if handler is None:
            return None

        method_name = handler_call_details.method
        custom_find_name = self._custom_find_name

        if not handler.request_streaming and not handler.response_streaming:
            handler_factory = grpc.unary_unary_rpc_method_handler

            async def wrapped(request: "Any", context: "ServicerContext") -> "Any":
                with sentry_sdk.isolation_scope():
                    name = (
                        custom_find_name(context) if custom_find_name else method_name
                    )
                    if not name:
                        return await handler(request, context)

                    span_streaming = has_span_streaming_enabled(
                        sentry_sdk.get_client().options
                    )
                    if span_streaming:
                        # What if the headers are empty?
                        sentry_sdk.traces.continue_trace(
                            dict(context.invocation_metadata())
                        )

                        with sentry_sdk.traces.start_span(
                            name=name,
                            attributes={
                                "sentry.op": OP.GRPC_SERVER,
                                "sentry.segment.name.source": SegmentNameSource.CUSTOM.value,
                                "sentry.origin": SPAN_ORIGIN,
                            },
                            parent_span=None,
                        ):
                            try:
                                return await handler.unary_unary(request, context)
                            except AbortError:
                                raise
                            except Exception as exc:
                                event, hint = event_from_exception(
                                    exc,
                                    mechanism={"type": "grpc", "handled": False},
                                )
                                sentry_sdk.capture_event(event, hint=hint)
                                raise
                    else:
                        # What if the headers are empty?
                        transaction = sentry_sdk.continue_trace(
                            dict(context.invocation_metadata()),
                            op=OP.GRPC_SERVER,
                            name=name,
                            source=TransactionSource.CUSTOM,
                            origin=SPAN_ORIGIN,
                        )

                        with sentry_sdk.start_transaction(transaction=transaction):
                            try:
                                return await handler.unary_unary(request, context)
                            except AbortError:
                                raise
                            except Exception as exc:
                                event, hint = event_from_exception(
                                    exc,
                                    mechanism={"type": "grpc", "handled": False},
                                )
                                sentry_sdk.capture_event(event, hint=hint)
                                raise

        elif not handler.request_streaming and handler.response_streaming:
            handler_factory = grpc.unary_stream_rpc_method_handler

            async def wrapped(request: "Any", context: "ServicerContext") -> "Any":  # type: ignore
                async for r in handler.unary_stream(request, context):
                    yield r

        elif handler.request_streaming and not handler.response_streaming:
            handler_factory = grpc.stream_unary_rpc_method_handler

            async def wrapped(request: "Any", context: "ServicerContext") -> "Any":
                response = handler.stream_unary(request, context)
                return await response

        elif handler.request_streaming and handler.response_streaming:
            handler_factory = grpc.stream_stream_rpc_method_handler

            async def wrapped(request: "Any", context: "ServicerContext") -> "Any":  # type: ignore
                async for r in handler.stream_stream(request, context):
                    yield r

        return handler_factory(
            wrapped,
            request_deserializer=handler.request_deserializer,
            response_serializer=handler.response_serializer,
        )



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import safe_serialize

from ..consts import SPAN_ORIGIN
from ..utils import _set_agent_data, _should_send_prompts

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from pydantic_ai._tool_manager import ToolDefinition  # type: ignore


def execute_tool_span(
    tool_name: str,
    tool_args: "Any",
    agent: "Any",
    tool_definition: "Optional[ToolDefinition]" = None,
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    """Create a span for tool execution.

    Args:
        tool_name: The name of the tool being executed
        tool_args: The arguments passed to the tool
        agent: The agent executing the tool
        tool_definition: The definition of the tool, if available
    """
    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"execute_tool {tool_name}",
            attributes={
                "sentry.op": OP.GEN_AI_EXECUTE_TOOL,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "execute_tool",
                SPANDATA.GEN_AI_TOOL_NAME: tool_name,
            },
        )

        set_on_span = span.set_attribute
    else:
        span = sentry_sdk.start_span(
            op=OP.GEN_AI_EXECUTE_TOOL,
            name=f"execute_tool {tool_name}",
            origin=SPAN_ORIGIN,
        )

        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "execute_tool")
        span.set_data(SPANDATA.GEN_AI_TOOL_NAME, tool_name)

        set_on_span = span.set_data

    if tool_definition is not None and hasattr(tool_definition, "description"):
        set_on_span(
            SPANDATA.GEN_AI_TOOL_DESCRIPTION,
            tool_definition.description,
        )

    _set_agent_data(span, agent)

    if _should_send_prompts() and tool_args is not None:
        set_on_span(SPANDATA.GEN_AI_TOOL_INPUT, safe_serialize(tool_args))

    return span


def update_execute_tool_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]", result: "Any"
) -> None:
    """Update the execute tool span with the result."""
    if not span:
        return

    if not _should_send_prompts() or result is None:
        return

    if isinstance(span, StreamedSpan):
        span.set_attribute(SPANDATA.GEN_AI_TOOL_OUTPUT, safe_serialize(result))
    else:
        span.set_data(SPANDATA.GEN_AI_TOOL_OUTPUT, safe_serialize(result))



import sys
import warnings
import weakref
from inspect import isawaitable
from typing import TYPE_CHECKING
from urllib.parse import urlsplit

import sentry_sdk
from sentry_sdk import continue_trace
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations._wsgi_common import RequestExtractor, _filter_headers
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SegmentNameSource, StreamedSpan
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    CONTEXTVARS_ERROR_MESSAGE,
    HAS_REAL_CONTEXTVARS,
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    parse_version,
    reraise,
)

if TYPE_CHECKING:
    from collections.abc import Container
    from typing import Any, Callable, Dict, Optional, Union

    from sanic.request import Request, RequestParameters
    from sanic.response import BaseHTTPResponse
    from sanic.router import Route

    from sentry_sdk._types import Event, EventProcessor, ExcInfo, Hint

try:
    from sanic import Sanic
    from sanic import __version__ as SANIC_VERSION
    from sanic.exceptions import SanicException
    from sanic.handlers import ErrorHandler
    from sanic.router import Router
except ImportError:
    raise DidNotEnable("Sanic not installed")

old_error_handler_lookup = ErrorHandler.lookup
old_handle_request = Sanic.handle_request
old_router_get = Router.get

try:
    # This method was introduced in Sanic v21.9
    old_startup = Sanic._startup
except AttributeError:
    pass


class SanicIntegration(Integration):
    identifier = "sanic"
    origin = f"auto.http.{identifier}"
    version: "Optional[tuple[int, ...]]" = None

    def __init__(
        self, unsampled_statuses: "Optional[Container[int]]" = frozenset({404})
    ) -> None:
        """
        The unsampled_statuses parameter can be used to specify for which HTTP statuses the
        transactions should not be sent to Sentry. By default, transactions are sent for all
        HTTP statuses, except 404. Set unsampled_statuses to None to send transactions for all
        HTTP statuses, including 404.
        """
        self._unsampled_statuses = unsampled_statuses or set()

    @staticmethod
    def setup_once() -> None:
        SanicIntegration.version = parse_version(SANIC_VERSION)
        _check_minimum_version(SanicIntegration, SanicIntegration.version)

        if not HAS_REAL_CONTEXTVARS:
            # We better have contextvars or we're going to leak state between
            # requests.
            raise DidNotEnable(
                "The sanic integration for Sentry requires Python 3.7+ "
                " or the aiocontextvars package." + CONTEXTVARS_ERROR_MESSAGE
            )

        if SANIC_VERSION.startswith("0.8."):
            # Sanic 0.8 and older creates a logger named "root" and puts a
            # stringified version of every exception in there (without exc_info),
            # which our error deduplication can't detect.
            #
            # We explicitly check the version here because it is a very
            # invasive step to ignore this logger and not necessary in newer
            # versions at all.
            #
            # https://github.com/huge-success/sanic/issues/1332
            ignore_logger("root")

        if SanicIntegration.version is not None and SanicIntegration.version < (21, 9):
            _setup_legacy_sanic()
            return

        _setup_sanic()


class SanicRequestExtractor(RequestExtractor):
    def content_length(self) -> int:
        if self.request.body is None:
            return 0
        return len(self.request.body)

    def cookies(self) -> "Dict[str, str]":
        return dict(self.request.cookies)

    def raw_data(self) -> bytes:
        return self.request.body

    def form(self) -> "RequestParameters":
        return self.request.form

    def is_json(self) -> bool:
        raise NotImplementedError()

    def json(self) -> "Optional[Any]":
        return self.request.json

    def files(self) -> "RequestParameters":
        return self.request.files

    def size_of_file(self, file: "Any") -> int:
        return len(file.body or ())


def _setup_sanic() -> None:
    Sanic._startup = _startup
    ErrorHandler.lookup = _sentry_error_handler_lookup


def _setup_legacy_sanic() -> None:
    Sanic.handle_request = _legacy_handle_request
    Router.get = _legacy_router_get
    ErrorHandler.lookup = _sentry_error_handler_lookup


async def _startup(self: "Sanic") -> None:
    # This happens about as early in the lifecycle as possible, just after the
    # Request object is created. The body has not yet been consumed.
    self.signal("http.lifecycle.request")(_context_enter)

    # This happens after the handler is complete. In v21.9 this signal is not
    # dispatched when there is an exception. Therefore we need to close out
    # and call _context_exit from the custom exception handler as well.
    # See https://github.com/sanic-org/sanic/issues/2297
    self.signal("http.lifecycle.response")(_context_exit)

    # This happens inside of request handling immediately after the route
    # has been identified by the router.
    self.signal("http.routing.after")(_set_transaction)

    # The above signals need to be declared before this can be called.
    await old_startup(self)


async def _context_enter(request: "Request") -> None:
    request.ctx._sentry_do_integration = (
        sentry_sdk.get_client().get_integration(SanicIntegration) is not None
    )

    if not request.ctx._sentry_do_integration:
        return

    client = sentry_sdk.get_client()
    is_span_streaming_enabled = has_span_streaming_enabled(client.options)

    weak_request = weakref.ref(request)
    request.ctx._sentry_scope = sentry_sdk.isolation_scope()
    scope = request.ctx._sentry_scope.__enter__()
    scope.clear_breadcrumbs()
    scope.add_event_processor(_make_request_processor(weak_request))

    if is_span_streaming_enabled:
        integration = client.get_integration(SanicIntegration)
        if (
            isinstance(integration, SanicIntegration)
            and integration._unsampled_statuses
        ):
            warnings.warn(
                "The `unsampled_statuses` option of SanicIntegration has no effect when span streaming is enabled.",
                stacklevel=2,
            )

        sentry_sdk.traces.continue_trace(dict(request.headers))
        scope.set_custom_sampling_context({"sanic_request": request})

        if should_send_default_pii() and request.remote_addr:
            scope.set_attribute(SPANDATA.USER_IP_ADDRESS, request.remote_addr)

        span = sentry_sdk.traces.start_span(
            # Unless the request results in a 404 error, the name and source
            # will get overwritten in _set_transaction
            name=request.path,
            attributes={
                "sentry.op": OP.HTTP_SERVER,
                "sentry.origin": SanicIntegration.origin,
                "sentry.segment.name.source": SegmentNameSource.URL.value,
            },
            parent_span=None,
        )
        request.ctx._sentry_root_span = span
    else:
        transaction = continue_trace(
            dict(request.headers),
            op=OP.HTTP_SERVER,
            # Unless the request results in a 404 error, the name and source will get overwritten in _set_transaction
            name=request.path,
            source=TransactionSource.URL,
            origin=SanicIntegration.origin,
        )
        request.ctx._sentry_root_span = sentry_sdk.start_transaction(
            transaction
        ).__enter__()


async def _context_exit(
    request: "Request", response: "Optional[BaseHTTPResponse]" = None
) -> None:
    with capture_internal_exceptions():
        if not request.ctx._sentry_do_integration:
            return

        integration = sentry_sdk.get_client().get_integration(SanicIntegration)

        response_status = None if response is None else response.status

        # This capture_internal_exceptions block has been intentionally nested here, so that in case an exception
        # happens while trying to end the transaction, we still attempt to exit the hub.
        with capture_internal_exceptions():
            span = request.ctx._sentry_root_span
            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    for attr, value in _get_request_attributes(request).items():
                        span.set_attribute(attr, value)
                if response_status is not None:
                    span.set_attribute(SPANDATA.HTTP_STATUS_CODE, response_status)
                    span.status = "error" if response_status >= 400 else "ok"

                span.end()

            else:
                span.set_http_status(response_status)
                span.sampled &= (
                    isinstance(integration, SanicIntegration)
                    and response_status not in integration._unsampled_statuses
                )

                span.__exit__(None, None, None)

        request.ctx._sentry_scope.__exit__(None, None, None)


async def _set_transaction(request: "Request", route: "Route", **_: "Any") -> None:
    if request.ctx._sentry_do_integration:
        with capture_internal_exceptions():
            scope = sentry_sdk.get_current_scope()
            route_name = route.name.replace(request.app.name, "").strip(".")
            scope.set_transaction_name(route_name, source=TransactionSource.COMPONENT)


def _sentry_error_handler_lookup(
    self: "Any", exception: Exception, *args: "Any", **kwargs: "Any"
) -> "Optional[object]":
    _capture_exception(exception)
    old_error_handler = old_error_handler_lookup(self, exception, *args, **kwargs)

    if old_error_handler is None:
        return None

    if sentry_sdk.get_client().get_integration(SanicIntegration) is None:
        return old_error_handler

    async def sentry_wrapped_error_handler(
        request: "Request", exception: Exception
    ) -> "Any":
        try:
            response = old_error_handler(request, exception)
            if isawaitable(response):
                response = await response
            return response
        except Exception:
            # Report errors that occur in Sanic error handler. These
            # exceptions will not even show up in Sanic's
            # `sanic.exceptions` logger.
            exc_info = sys.exc_info()
            _capture_exception(exc_info)
            reraise(*exc_info)
        finally:
            # As mentioned in previous comment in _startup, this can be removed
            # after https://github.com/sanic-org/sanic/issues/2297 is resolved
            if SanicIntegration.version and SanicIntegration.version == (21, 9):
                await _context_exit(request)

    return sentry_wrapped_error_handler


async def _legacy_handle_request(
    self: "Any", request: "Request", *args: "Any", **kwargs: "Any"
) -> "Any":
    if sentry_sdk.get_client().get_integration(SanicIntegration) is None:
        return await old_handle_request(self, request, *args, **kwargs)

    weak_request = weakref.ref(request)

    with sentry_sdk.isolation_scope() as scope:
        scope.clear_breadcrumbs()
        scope.add_event_processor(_make_request_processor(weak_request))

        response = old_handle_request(self, request, *args, **kwargs)
        if isawaitable(response):
            response = await response

        return response


def _legacy_router_get(self: "Any", *args: "Union[Any, Request]") -> "Any":
    rv = old_router_get(self, *args)
    if sentry_sdk.get_client().get_integration(SanicIntegration) is not None:
        with capture_internal_exceptions():
            scope = sentry_sdk.get_isolation_scope()
            if SanicIntegration.version and SanicIntegration.version >= (21, 3):
                # Sanic versions above and including 21.3 append the app name to the
                # route name, and so we need to remove it from Route name so the
                # transaction name is consistent across all versions
                sanic_app_name = self.ctx.app.name
                sanic_route = rv[0].name

                if sanic_route.startswith("%s." % sanic_app_name):
                    # We add a 1 to the len of the sanic_app_name because there is a dot
                    # that joins app name and the route name
                    # Format: app_name.route_name
                    sanic_route = sanic_route[len(sanic_app_name) + 1 :]

                scope.set_transaction_name(
                    sanic_route, source=TransactionSource.COMPONENT
                )
            else:
                scope.set_transaction_name(
                    rv[0].__name__, source=TransactionSource.COMPONENT
                )

    return rv


@ensure_integration_enabled(SanicIntegration)
def _capture_exception(exception: "Union[ExcInfo, BaseException]") -> None:
    with capture_internal_exceptions():
        event, hint = event_from_exception(
            exception,
            client_options=sentry_sdk.get_client().options,
            mechanism={"type": "sanic", "handled": False},
        )

        if hint and hasattr(hint["exc_info"][0], "quiet") and hint["exc_info"][0].quiet:
            return

        sentry_sdk.capture_event(event, hint=hint)


def _get_request_attributes(request: "Request") -> "Dict[str, Any]":
    """
    Return span attributes related to the HTTP request from a Sanic request.
    """
    attributes = {}  # type: Dict[str, Any]

    if request.method:
        attributes[SPANDATA.HTTP_REQUEST_METHOD] = request.method.upper()

    headers = _filter_headers(dict(request.headers), use_annotated_value=False)
    for header, value in headers.items():
        attributes[f"{SPANDATA.HTTP_REQUEST_HEADER}.{header.lower()}"] = value

    urlparts = urlsplit(request.url)

    if should_send_default_pii():
        attributes[SPANDATA.URL_FULL] = request.url
        attributes["url.path"] = urlparts.path

        if urlparts.query:
            attributes[SPANDATA.HTTP_QUERY] = urlparts.query

    if urlparts.scheme:
        attributes[SPANDATA.NETWORK_PROTOCOL_NAME] = urlparts.scheme

    if should_send_default_pii() and request.remote_addr:
        attributes[SPANDATA.CLIENT_ADDRESS] = request.remote_addr

    return attributes


def _make_request_processor(weak_request: "Callable[[], Request]") -> "EventProcessor":
    def sanic_processor(event: "Event", hint: "Optional[Hint]") -> "Optional[Event]":
        try:
            if hint and issubclass(hint["exc_info"][0], SanicException):
                return None
        except KeyError:
            pass

        request = weak_request()
        if request is None:
            return event

        with capture_internal_exceptions():
            extractor = SanicRequestExtractor(request)
            extractor.extract_into_event(event)

            request_info = event["request"]
            urlparts = urlsplit(request.url)

            request_info["url"] = "%s://%s%s" % (
                urlparts.scheme,
                urlparts.netloc,
                urlparts.path,
            )

            request_info["query_string"] = urlparts.query
            request_info["method"] = request.method
            request_info["env"] = {"REMOTE_ADDR": request.remote_addr}
            request_info["headers"] = _filter_headers(dict(request.headers))

        return event

    return sanic_processor



from typing import TYPE_CHECKING, Dict, List, cast

from sentry_sdk.utils import (
    AnnotatedValue,
    capture_internal_exceptions,
    iter_event_frames,
)

if TYPE_CHECKING:
    from typing import Optional

    from sentry_sdk._types import Event


DEFAULT_DENYLIST = [
    # stolen from relay
    "password",
    "passwd",
    "secret",
    "api_key",
    "apikey",
    "auth",
    "credentials",
    "mysql_pwd",
    "privatekey",
    "private_key",
    "token",
    "session",
    # django
    "csrftoken",
    "sessionid",
    # wsgi
    "x_csrftoken",
    "x_forwarded_for",
    "set_cookie",
    "cookie",
    "authorization",
    "proxy-authorization",
    "x_api_key",
    # other common names used in the wild
    "aiohttp_session",  # aiohttp
    "connect.sid",  # Express
    "csrf_token",  # Pyramid
    "csrf",  # (this is a cookie name used in accepted answers on stack overflow)
    "_csrf",  # Express
    "_csrf_token",  # Bottle
    "PHPSESSID",  # PHP
    "_session",  # Sanic
    "symfony",  # Symfony
    "user_session",  # Vue
    "_xsrf",  # Tornado
    "XSRF-TOKEN",  # Angular, Laravel
]

DEFAULT_PII_DENYLIST = [
    "x_forwarded_for",
    "x_real_ip",
    "ip_address",
    "remote_addr",
]


class EventScrubber:
    def __init__(
        self,
        denylist: "Optional[List[str]]" = None,
        recursive: bool = False,
        send_default_pii: bool = False,
        pii_denylist: "Optional[List[str]]" = None,
    ) -> None:
        """
        A scrubber that goes through the event payload and removes sensitive data configured through denylists.

        :param denylist: A security denylist that is always scrubbed, defaults to DEFAULT_DENYLIST.
        :param recursive: Whether to scrub the event payload recursively, default False.
        :param send_default_pii: Whether pii is sending is on, pii fields are not scrubbed.
        :param pii_denylist: The denylist to use for scrubbing when pii is not sent, defaults to DEFAULT_PII_DENYLIST.
        """
        self.denylist = DEFAULT_DENYLIST.copy() if denylist is None else denylist

        if not send_default_pii:
            pii_denylist = (
                DEFAULT_PII_DENYLIST.copy() if pii_denylist is None else pii_denylist
            )
            self.denylist += pii_denylist

        self.denylist = [x.lower() for x in self.denylist]
        self.recursive = recursive

    def scrub_list(self, lst: object) -> None:
        """
        If a list is passed to this method, the method recursively searches the list and any
        nested lists for any dictionaries. The method calls scrub_dict on all dictionaries
        it finds.
        If the parameter passed to this method is not a list, the method does nothing.
        """
        if not isinstance(lst, list):
            return

        for v in lst:
            self.scrub_dict(v)  # no-op unless v is a dict
            self.scrub_list(v)  # no-op unless v is a list

    def scrub_dict(self, d: object) -> None:
        """
        If a dictionary is passed to this method, the method scrubs the dictionary of any
        sensitive data. The method calls itself recursively on any nested dictionaries (
        including dictionaries nested in lists) if self.recursive is True.
        This method does nothing if the parameter passed to it is not a dictionary.
        """
        if not isinstance(d, dict):
            return

        for k, v in d.items():
            # The cast is needed because mypy is not smart enough to figure out that k must be a
            # string after the isinstance check.
            if isinstance(k, str) and k.lower() in self.denylist:
                d[k] = AnnotatedValue.substituted_because_contains_sensitive_data()
            elif self.recursive:
                self.scrub_dict(v)  # no-op unless v is a dict
                self.scrub_list(v)  # no-op unless v is a list

    def scrub_request(self, event: "Event") -> None:
        with capture_internal_exceptions():
            if "request" in event:
                if "headers" in event["request"]:
                    self.scrub_dict(event["request"]["headers"])
                if "cookies" in event["request"]:
                    self.scrub_dict(event["request"]["cookies"])
                if "data" in event["request"]:
                    self.scrub_dict(event["request"]["data"])

    def scrub_extra(self, event: "Event") -> None:
        with capture_internal_exceptions():
            if "extra" in event:
                self.scrub_dict(event["extra"])

    def scrub_user(self, event: "Event") -> None:
        with capture_internal_exceptions():
            if "user" in event:
                user = event["user"]
                if "ip_address" in self.denylist and isinstance(user, dict):
                    user.pop("ip_address", None)
                self.scrub_dict(user)

    def scrub_breadcrumbs(self, event: "Event") -> None:
        with capture_internal_exceptions():
            if "breadcrumbs" in event:
                if (
                    not isinstance(event["breadcrumbs"], AnnotatedValue)
                    and "values" in event["breadcrumbs"]
                ):
                    for value in event["breadcrumbs"]["values"]:
                        if "data" in value:
                            self.scrub_dict(value["data"])

    def scrub_frames(self, event: "Event") -> None:
        with capture_internal_exceptions():
            for frame in iter_event_frames(event):
                if "vars" in frame:
                    self.scrub_dict(frame["vars"])

    def scrub_spans(self, event: "Event") -> None:
        with capture_internal_exceptions():
            if "spans" in event:
                for span in cast(List[Dict[str, object]], event["spans"]):
                    if "data" in span:
                        self.scrub_dict(span["data"])

    def scrub_event(self, event: "Event") -> None:
        self.scrub_request(event)
        self.scrub_extra(event)
        self.scrub_user(event)
        self.scrub_breadcrumbs(event)
        self.scrub_frames(event)
        self.scrub_spans(event)



from typing import TYPE_CHECKING

from sentry_sdk.feature_flags import add_feature_flag
from sentry_sdk.integrations import DidNotEnable, Integration

try:
    import ldclient
    from ldclient.hook import Hook, Metadata

    if TYPE_CHECKING:
        from typing import Any

        from ldclient import LDClient
        from ldclient.evaluation import EvaluationDetail
        from ldclient.hook import EvaluationSeriesContext
except ImportError:
    raise DidNotEnable("LaunchDarkly is not installed")


class LaunchDarklyIntegration(Integration):
    identifier = "launchdarkly"

    def __init__(self, ld_client: "LDClient | None" = None) -> None:
        """
        :param client: An initialized LDClient instance. If a client is not provided, this
            integration will attempt to use the shared global instance.
        """
        try:
            client = ld_client or ldclient.get()
        except Exception as exc:
            raise DidNotEnable("Error getting LaunchDarkly client. " + repr(exc))

        if not client.is_initialized():
            raise DidNotEnable("LaunchDarkly client is not initialized.")

        # Register the flag collection hook with the LD client.
        client.add_hook(LaunchDarklyHook())

    @staticmethod
    def setup_once() -> None:
        pass


class LaunchDarklyHook(Hook):
    @property
    def metadata(self) -> "Metadata":
        return Metadata(name="sentry-flag-auditor")

    def after_evaluation(
        self,
        series_context: "EvaluationSeriesContext",
        data: "dict[Any, Any]",
        detail: "EvaluationDetail",
    ) -> "dict[Any, Any]":
        if isinstance(detail.value, bool):
            add_feature_flag(series_context.key, detail.value)

        return data

    def before_evaluation(
        self, series_context: "EvaluationSeriesContext", data: "dict[Any, Any]"
    ) -> "dict[Any, Any]":
        return data  # No-op.



import sys
from collections.abc import Mapping
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk import isolation_scope
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP, SPANDATA, SPANSTATUS
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations.celery.beat import (
    _patch_beat_apply_entry,
    _patch_redbeat_apply_async,
    _setup_celery_beat_signals,
)
from sentry_sdk.integrations.celery.utils import _now_seconds_since_epoch
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.scope import Scope, should_send_default_pii
from sentry_sdk.traces import SegmentNameSource, StreamedSpan, get_current_span
from sentry_sdk.tracing import BAGGAGE_HEADER_NAME, Span, TransactionSource
from sentry_sdk.tracing_utils import Baggage, has_span_streaming_enabled
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    capture_internal_exceptions,
    event_from_exception,
    reraise,
)

if TYPE_CHECKING:
    from typing import Any, Callable, List, Optional, TypeVar, Union

    from sentry_sdk._types import Event, EventProcessor, ExcInfo, Hint

    F = TypeVar("F", bound=Callable[..., Any])


try:
    from celery import VERSION as CELERY_VERSION  # type: ignore
    from celery.app.task import Task  # type: ignore
    from celery.app.trace import task_has_custom
    from celery.exceptions import (  # type: ignore
        Ignore,
        Reject,
        Retry,
        SoftTimeLimitExceeded,
    )
    from kombu import Producer  # type: ignore
except ImportError:
    raise DidNotEnable("Celery not installed")


CELERY_CONTROL_FLOW_EXCEPTIONS = (Retry, Ignore, Reject)


class CeleryIntegration(Integration):
    identifier = "celery"
    origin = f"auto.queue.{identifier}"

    def __init__(
        self,
        propagate_traces: bool = True,
        monitor_beat_tasks: bool = False,
        exclude_beat_tasks: "Optional[List[str]]" = None,
    ) -> None:
        self.propagate_traces = propagate_traces
        self.monitor_beat_tasks = monitor_beat_tasks
        self.exclude_beat_tasks = exclude_beat_tasks

        _patch_beat_apply_entry()
        _patch_redbeat_apply_async()
        _setup_celery_beat_signals(monitor_beat_tasks)

    @staticmethod
    def setup_once() -> None:
        _check_minimum_version(CeleryIntegration, CELERY_VERSION)

        _patch_build_tracer()
        _patch_task_apply_async()
        _patch_celery_send_task()
        _patch_worker_exit()
        _patch_producer_publish()

        # This logger logs every status of every task that ran on the worker.
        # Meaning that every task's breadcrumbs are full of stuff like "Task
        # <foo> raised unexpected <bar>".
        ignore_logger("celery.worker.job")
        ignore_logger("celery.app.trace")

        # This is stdout/err redirected to a logger, can't deal with this
        # (need event_level=logging.WARN to reproduce)
        ignore_logger("celery.redirected")


def _set_status(status: str) -> None:
    client = sentry_sdk.get_client()
    span_streaming = has_span_streaming_enabled(client.options)

    with capture_internal_exceptions():
        scope = sentry_sdk.get_current_scope()

        if span_streaming and scope.streamed_span is not None:
            scope.streamed_span.status = "ok" if status == "ok" else "error"
        elif not span_streaming and scope.span is not None:
            scope.span.set_status(status)


def _capture_exception(task: "Any", exc_info: "ExcInfo") -> None:
    client = sentry_sdk.get_client()
    if client.get_integration(CeleryIntegration) is None:
        return

    if isinstance(exc_info[1], CELERY_CONTROL_FLOW_EXCEPTIONS):
        # ??? Doesn't map to anything
        _set_status("aborted")
        return

    _set_status("internal_error")

    if hasattr(task, "throws") and isinstance(exc_info[1], task.throws):
        return

    event, hint = event_from_exception(
        exc_info,
        client_options=client.options,
        mechanism={"type": "celery", "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)


def _make_event_processor(
    task: "Any",
    uuid: "Any",
    args: "Any",
    kwargs: "Any",
    request: "Optional[Any]" = None,
) -> "EventProcessor":
    def event_processor(event: "Event", hint: "Hint") -> "Optional[Event]":
        with capture_internal_exceptions():
            tags = event.setdefault("tags", {})
            tags["celery_task_id"] = uuid
            extra = event.setdefault("extra", {})
            extra["celery-job"] = {
                "task_name": task.name,
                "args": (
                    args if should_send_default_pii() else SENSITIVE_DATA_SUBSTITUTE
                ),
                "kwargs": (
                    kwargs if should_send_default_pii() else SENSITIVE_DATA_SUBSTITUTE
                ),
            }

        if "exc_info" in hint:
            with capture_internal_exceptions():
                if issubclass(hint["exc_info"][0], SoftTimeLimitExceeded):
                    event["fingerprint"] = [
                        "celery",
                        "SoftTimeLimitExceeded",
                        getattr(task, "name", task),
                    ]

        return event

    return event_processor


def _update_celery_task_headers(
    original_headers: "dict[str, Any]",
    span: "Optional[Union[StreamedSpan, Span]]",
    monitor_beat_tasks: bool,
) -> "dict[str, Any]":
    """
    Updates the headers of the Celery task with the tracing information
    and eventually Sentry Crons monitoring information for beat tasks.
    """
    updated_headers = original_headers.copy()
    with capture_internal_exceptions():
        # if span is None (when the task was started by Celery Beat)
        # this will return the trace headers from the scope.
        headers = dict(
            sentry_sdk.get_isolation_scope().iter_trace_propagation_headers(span=span)
        )

        if monitor_beat_tasks:
            headers.update(
                {
                    "sentry-monitor-start-timestamp-s": "%.9f"
                    % _now_seconds_since_epoch(),
                }
            )

        # Add the time the task was enqueued to the headers
        # This is used in the consumer to calculate the latency
        updated_headers.update(
            {"sentry-task-enqueued-time": _now_seconds_since_epoch()}
        )

        if headers:
            existing_baggage = updated_headers.get(BAGGAGE_HEADER_NAME)
            sentry_baggage = headers.get(BAGGAGE_HEADER_NAME)

            combined_baggage = sentry_baggage or existing_baggage
            if sentry_baggage and existing_baggage:
                # Merge incoming and sentry baggage, where the sentry trace information
                # in the incoming baggage takes precedence and the third-party items
                # are concatenated.
                incoming = Baggage.from_incoming_header(existing_baggage)
                combined = Baggage.from_incoming_header(sentry_baggage)
                combined.sentry_items.update(incoming.sentry_items)
                combined.third_party_items = ",".join(
                    [
                        x
                        for x in [
                            combined.third_party_items,
                            incoming.third_party_items,
                        ]
                        if x is not None and x != ""
                    ]
                )
                combined_baggage = combined.serialize(include_third_party=True)

            updated_headers.update(headers)
            if combined_baggage:
                updated_headers[BAGGAGE_HEADER_NAME] = combined_baggage

            # https://github.com/celery/celery/issues/4875
            #
            # Need to setdefault the inner headers too since other
            # tracing tools (dd-trace-py) also employ this exact
            # workaround and we don't want to break them.
            updated_headers.setdefault("headers", {}).update(headers)
            if combined_baggage:
                updated_headers["headers"][BAGGAGE_HEADER_NAME] = combined_baggage

            # Add the Sentry options potentially added in `sentry_apply_entry`
            # to the headers (done when auto-instrumenting Celery Beat tasks)
            for key, value in updated_headers.items():
                if key.startswith("sentry-"):
                    updated_headers["headers"][key] = value

            # Preserve user-provided custom headers in the inner "headers" dict
            # so they survive to task.request.headers on the worker (celery#4875).
            for key, value in original_headers.items():
                if key != "headers" and key not in updated_headers["headers"]:
                    updated_headers["headers"][key] = value

    return updated_headers


class NoOpMgr:
    def __enter__(self) -> None:
        return None

    def __exit__(self, exc_type: "Any", exc_value: "Any", traceback: "Any") -> None:
        return None


def _wrap_task_run(f: "F") -> "F":
    @wraps(f)
    def apply_async(*args: "Any", **kwargs: "Any") -> "Any":
        # Note: kwargs can contain headers=None, so no setdefault!
        # Unsure which backend though.
        client = sentry_sdk.get_client()
        integration = client.get_integration(CeleryIntegration)
        if integration is None:
            return f(*args, **kwargs)

        kwarg_headers = kwargs.get("headers") or {}
        propagate_traces = kwarg_headers.pop(
            "sentry-propagate-traces", integration.propagate_traces
        )

        if not propagate_traces:
            return f(*args, **kwargs)

        if isinstance(args[0], Task):
            task_name: str = args[0].name
        elif len(args) > 1 and isinstance(args[1], str):
            task_name = args[1]
        else:
            task_name = "<unknown Celery task>"

        span_streaming = has_span_streaming_enabled(client.options)

        task_started_from_beat = sentry_sdk.get_isolation_scope()._name == "celery-beat"

        span_mgr: "Union[StreamedSpan, Span, NoOpMgr]" = NoOpMgr()
        if span_streaming:
            if not task_started_from_beat and get_current_span() is not None:
                span_mgr = sentry_sdk.traces.start_span(
                    name=task_name,
                    attributes={
                        "sentry.op": OP.QUEUE_SUBMIT_CELERY,
                        "sentry.origin": CeleryIntegration.origin,
                    },
                )

        else:
            if not task_started_from_beat:
                span_mgr = sentry_sdk.start_span(
                    op=OP.QUEUE_SUBMIT_CELERY,
                    name=task_name,
                    origin=CeleryIntegration.origin,
                )

        with span_mgr as span:
            kwargs["headers"] = _update_celery_task_headers(
                kwarg_headers, span, integration.monitor_beat_tasks
            )
            return f(*args, **kwargs)

    return apply_async  # type: ignore


def _wrap_tracer(task: "Any", f: "F") -> "F":
    # Need to wrap tracer for pushing the scope before prerun is sent, and
    # popping it after postrun is sent.
    #
    # This is the reason we don't use signals for hooking in the first place.
    # Also because in Celery 3, signal dispatch returns early if one handler
    # crashes.
    @wraps(f)
    def _inner(*args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        if client.get_integration(CeleryIntegration) is None:
            return f(*args, **kwargs)

        span_streaming = has_span_streaming_enabled(client.options)

        with isolation_scope() as scope:
            scope._name = "celery"
            scope.clear_breadcrumbs()
            scope.add_event_processor(_make_event_processor(task, *args, **kwargs))

            task_name = getattr(task, "name", "<unknown Celery task>")

            custom_sampling_context = {}
            with capture_internal_exceptions():
                custom_sampling_context = {
                    "celery_job": {
                        "task": task_name,
                        # for some reason, args[1] is a list if non-empty but a
                        # tuple if empty
                        "args": list(args[1]),
                        "kwargs": args[2],
                    }
                }

            span: "Union[Span, StreamedSpan]"
            span_ctx: "Union[StreamedSpan, Span, NoOpMgr]" = NoOpMgr()

            # Celery task objects are not a thing to be trusted. Even
            # something such as attribute access can fail.
            with capture_internal_exceptions():
                headers = args[3].get("headers") or {}
                if span_streaming:
                    sentry_sdk.traces.continue_trace(headers)
                    Scope.set_custom_sampling_context(custom_sampling_context)
                    span = sentry_sdk.traces.start_span(
                        name=task_name,
                        parent_span=None,  # make this a segment
                        attributes={
                            "sentry.origin": CeleryIntegration.origin,
                            "sentry.segment.name.source": SegmentNameSource.TASK.value,
                            "sentry.op": OP.QUEUE_TASK_CELERY,
                        },
                    )

                    span_ctx = span

                else:
                    span = continue_trace(
                        headers,
                        op=OP.QUEUE_TASK_CELERY,
                        name=task_name,
                        source=TransactionSource.TASK,
                        origin=CeleryIntegration.origin,
                    )
                    span.set_status(SPANSTATUS.OK)

                    span_ctx = sentry_sdk.start_transaction(
                        span,
                        custom_sampling_context=custom_sampling_context,
                    )

            with span_ctx:
                return f(*args, **kwargs)

    return _inner  # type: ignore


def _set_messaging_destination_name(
    task: "Any", span: "Union[StreamedSpan, Span]"
) -> None:
    """Set "messaging.destination.name" tag for span"""
    with capture_internal_exceptions():
        delivery_info = task.request.delivery_info
        if delivery_info:
            routing_key = delivery_info.get("routing_key")
            if delivery_info.get("exchange") == "" and routing_key is not None:
                # Empty exchange indicates the default exchange, meaning the tasks
                # are sent to the queue with the same name as the routing key.
                if isinstance(span, StreamedSpan):
                    span.set_attribute(SPANDATA.MESSAGING_DESTINATION_NAME, routing_key)
                else:
                    span.set_data(SPANDATA.MESSAGING_DESTINATION_NAME, routing_key)


def _wrap_task_call(task: "Any", f: "F") -> "F":
    # Need to wrap task call because the exception is caught before we get to
    # see it. Also celery's reported stacktrace is untrustworthy.

    @wraps(f)
    def _inner(*args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        if client.get_integration(CeleryIntegration) is None:
            return f(*args, **kwargs)

        span_streaming = has_span_streaming_enabled(client.options)

        try:
            if span_streaming and get_current_span() is None:
                return f(*args, **kwargs)

            span: "Union[Span, StreamedSpan]"
            if span_streaming:
                span = sentry_sdk.traces.start_span(
                    name=task.name,
                    attributes={
                        "sentry.op": OP.QUEUE_PROCESS,
                        "sentry.origin": CeleryIntegration.origin,
                    },
                )
            else:
                span = sentry_sdk.start_span(
                    op=OP.QUEUE_PROCESS,
                    name=task.name,
                    origin=CeleryIntegration.origin,
                )

            with span:
                if isinstance(span, StreamedSpan):
                    set_on_span = span.set_attribute
                else:
                    set_on_span = span.set_data

                _set_messaging_destination_name(task, span)

                latency = None
                with capture_internal_exceptions():
                    if (
                        task.request.headers is not None
                        and "sentry-task-enqueued-time" in task.request.headers
                    ):
                        latency = _now_seconds_since_epoch() - task.request.headers.pop(
                            "sentry-task-enqueued-time"
                        )

                if latency is not None:
                    latency *= 1000  # milliseconds
                    set_on_span(SPANDATA.MESSAGING_MESSAGE_RECEIVE_LATENCY, latency)

                with capture_internal_exceptions():
                    set_on_span(SPANDATA.MESSAGING_MESSAGE_ID, task.request.id)

                with capture_internal_exceptions():
                    set_on_span(
                        SPANDATA.MESSAGING_MESSAGE_RETRY_COUNT,
                        task.request.retries,
                    )

                with capture_internal_exceptions():
                    with task.app.connection() as conn:
                        set_on_span(
                            SPANDATA.MESSAGING_SYSTEM,
                            conn.transport.driver_type,
                        )

                return f(*args, **kwargs)
        except Exception:
            exc_info = sys.exc_info()
            with capture_internal_exceptions():
                _capture_exception(task, exc_info)
            reraise(*exc_info)

    return _inner  # type: ignore


def _patch_build_tracer() -> None:
    import celery.app.trace as trace  # type: ignore

    original_build_tracer = trace.build_tracer

    def sentry_build_tracer(
        name: "Any", task: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        if not getattr(task, "_sentry_is_patched", False):
            # determine whether Celery will use __call__ or run and patch
            # accordingly
            if task_has_custom(task, "__call__"):
                type(task).__call__ = _wrap_task_call(task, type(task).__call__)
            else:
                task.run = _wrap_task_call(task, task.run)

            # `build_tracer` is apparently called for every task
            # invocation. Can't wrap every celery task for every invocation
            # or we will get infinitely nested wrapper functions.
            task._sentry_is_patched = True

        return _wrap_tracer(task, original_build_tracer(name, task, *args, **kwargs))

    trace.build_tracer = sentry_build_tracer


def _patch_task_apply_async() -> None:
    Task.apply_async = _wrap_task_run(Task.apply_async)


def _patch_celery_send_task() -> None:
    from celery import Celery

    Celery.send_task = _wrap_task_run(Celery.send_task)


def _patch_worker_exit() -> None:
    # Need to flush queue before worker shutdown because a crashing worker will
    # call os._exit
    from billiard.pool import Worker  # type: ignore

    original_workloop = Worker.workloop

    def sentry_workloop(*args: "Any", **kwargs: "Any") -> "Any":
        try:
            return original_workloop(*args, **kwargs)
        finally:
            with capture_internal_exceptions():
                if (
                    sentry_sdk.get_client().get_integration(CeleryIntegration)
                    is not None
                ):
                    sentry_sdk.flush()

    Worker.workloop = sentry_workloop


def _patch_producer_publish() -> None:
    original_publish = Producer.publish

    def sentry_publish(self: "Producer", *args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()
        if client.get_integration(CeleryIntegration) is None:
            return original_publish(self, *args, **kwargs)

        span_streaming = has_span_streaming_enabled(client.options)

        kwargs_headers = kwargs.get("headers", {})
        if not isinstance(kwargs_headers, Mapping):
            # Ensure kwargs_headers is a Mapping, so we can safely call get().
            # We don't expect this to happen, but it's better to be safe. Even
            # if it does happen, only our instrumentation breaks. This line
            # does not overwrite kwargs["headers"], so the original publish
            # method will still work.
            kwargs_headers = {}

        task_name = kwargs_headers.get("task") or "<unknown Celery task>"
        task_id = kwargs_headers.get("id")
        retries = kwargs_headers.get("retries")

        routing_key = kwargs.get("routing_key")
        exchange = kwargs.get("exchange")

        span: "Union[StreamedSpan, Span, None]" = None
        if span_streaming:
            if get_current_span() is not None:
                span = sentry_sdk.traces.start_span(
                    name=task_name,
                    attributes={
                        "sentry.op": OP.QUEUE_PUBLISH,
                        "sentry.origin": CeleryIntegration.origin,
                    },
                )
        else:
            span = sentry_sdk.start_span(
                op=OP.QUEUE_PUBLISH,
                name=task_name,
                origin=CeleryIntegration.origin,
            )

        if span is None:
            return original_publish(self, *args, **kwargs)

        with span:
            if isinstance(span, StreamedSpan):
                set_on_span = span.set_attribute
            else:
                set_on_span = span.set_data

            if task_id is not None:
                set_on_span(SPANDATA.MESSAGING_MESSAGE_ID, task_id)

            if exchange == "" and routing_key is not None:
                # Empty exchange indicates the default exchange, meaning messages are
                # routed to the queue with the same name as the routing key.
                set_on_span(SPANDATA.MESSAGING_DESTINATION_NAME, routing_key)

            if retries is not None:
                set_on_span(SPANDATA.MESSAGING_MESSAGE_RETRY_COUNT, retries)

            with capture_internal_exceptions():
                set_on_span(
                    SPANDATA.MESSAGING_SYSTEM, self.connection.transport.driver_type
                )

            return original_publish(self, *args, **kwargs)

    Producer.publish = sentry_publish



import re

# Matches data URLs with base64-encoded content, e.g. "data:image/png;base64,iVBORw0K..."
DATA_URL_BASE64_REGEX = re.compile(
    r"^data:(?:[a-zA-Z0-9][a-zA-Z0-9.+\-]*/[a-zA-Z0-9][a-zA-Z0-9.+\-]*)(?:;[a-zA-Z0-9\-]+=[^;,]*)*;base64,(?:[A-Za-z0-9+/\-_]+={0,2})$"
)



from .ai_client import ai_client_span, update_ai_client_span  # noqa: F401
from .execute_tool import execute_tool_span, update_execute_tool_span  # noqa: F401
from .invoke_agent import invoke_agent_span, update_invoke_agent_span  # noqa: F401



from sentry_sdk.profiler.continuous_profiler import (
    start_profile_session,
    start_profiler,
    stop_profile_session,
    stop_profiler,
)
from sentry_sdk.profiler.transaction_profiler import (
    MAX_PROFILE_DURATION_NS,
    PROFILE_MINIMUM_SAMPLES,
    GeventScheduler,
    Profile,
    Scheduler,
    ThreadScheduler,
    has_profiling_enabled,
    setup_profiler,
    teardown_profiler,
)
from sentry_sdk.profiler.utils import (
    DEFAULT_SAMPLING_FREQUENCY,
    MAX_STACK_DEPTH,
    extract_frame,
    extract_stack,
    frame_id,
    get_frame_name,
)

__all__ = [
    "start_profile_session",  # TODO: Deprecate this in favor of `start_profiler`
    "start_profiler",
    "stop_profile_session",  # TODO: Deprecate this in favor of `stop_profiler`
    "stop_profiler",
    # DEPRECATED: The following was re-exported for backwards compatibility. It
    # will be removed from sentry_sdk.profiler in a future release.
    "MAX_PROFILE_DURATION_NS",
    "PROFILE_MINIMUM_SAMPLES",
    "Profile",
    "Scheduler",
    "ThreadScheduler",
    "GeventScheduler",
    "has_profiling_enabled",
    "setup_profiler",
    "teardown_profiler",
    "DEFAULT_SAMPLING_FREQUENCY",
    "MAX_STACK_DEPTH",
    "get_frame_name",
    "extract_frame",
    "extract_stack",
    "frame_id",
]



import urllib
from enum import Enum
from typing import TYPE_CHECKING

from sentry_sdk.integrations._wsgi_common import _filter_headers
from sentry_sdk.scope import should_send_default_pii

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union

    from typing_extensions import Literal

    from sentry_sdk.utils import AnnotatedValue


class _RootPathInPath(Enum):
    EXCLUDED = "excluded"
    EITHER = "either"


def _get_headers(asgi_scope: "Any") -> "Dict[str, str]":
    """
    Extract headers from the ASGI scope, in the format that the Sentry protocol expects.
    """
    headers: "Dict[str, str]" = {}
    for raw_key, raw_value in asgi_scope["headers"]:
        key = raw_key.decode("latin-1")
        value = raw_value.decode("latin-1")
        if key in headers:
            headers[key] = headers[key] + ", " + value
        else:
            headers[key] = value

    return headers


def _get_path(
    asgi_scope: "Dict[str, Any]", root_path_in_path: "_RootPathInPath"
) -> "str":
    if root_path_in_path is _RootPathInPath.EXCLUDED:
        return asgi_scope.get("root_path", "") + asgi_scope.get("path", "")

    # Inverse of https://github.com/Kludex/starlette/blob/de970d7b3facb853eb7ad077decbf3d94f2aab6c/starlette/_utils.py#L96
    path = asgi_scope["path"]
    root_path = asgi_scope.get("root_path", "")

    if not root_path or path == root_path or path.startswith(root_path + "/"):
        return path

    return root_path + path


def _get_url(
    asgi_scope: "Dict[str, Any]",
    default_scheme: "Literal['ws', 'http']",
    host: "Optional[Union[AnnotatedValue, str]]",
    path: str,
) -> str:
    """
    Extract URL from the ASGI scope, without also including the querystring.
    """
    scheme = asgi_scope.get("scheme", default_scheme)

    server = asgi_scope.get("server", None)

    if host:
        return "%s://%s%s" % (scheme, host, path)

    if server is not None:
        host, port = server
        default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}.get(scheme)
        if port != default_port:
            return "%s://%s:%s%s" % (scheme, host, port, path)
        return "%s://%s%s" % (scheme, host, path)
    return path


def _get_query(asgi_scope: "Any") -> "Any":
    """
    Extract querystring from the ASGI scope, in the format that the Sentry protocol expects.
    """
    qs = asgi_scope.get("query_string")
    if not qs:
        return None
    return urllib.parse.unquote(qs.decode("latin-1"))


def _get_ip(asgi_scope: "Any") -> str:
    """
    Extract IP Address from the ASGI scope based on request headers with fallback to scope client.
    """
    headers = _get_headers(asgi_scope)
    try:
        return headers["x-forwarded-for"].split(",")[0].strip()
    except (KeyError, IndexError):
        pass

    try:
        return headers["x-real-ip"]
    except KeyError:
        pass

    return asgi_scope.get("client")[0]


def _get_request_data(
    asgi_scope: "Any",
    root_path_in_path: "_RootPathInPath",
) -> "Dict[str, Any]":
    """
    Returns data related to the HTTP request from the ASGI scope.
    """
    request_data: "Dict[str, Any]" = {}
    ty = asgi_scope["type"]
    if ty in ("http", "websocket"):
        request_data["method"] = asgi_scope.get("method")

        request_data["headers"] = headers = _filter_headers(
            _get_headers(asgi_scope),
        )
        request_data["query_string"] = _get_query(asgi_scope)

        request_data["url"] = _get_url(
            asgi_scope,
            "http" if ty == "http" else "ws",
            headers.get("host"),
            path=_get_path(asgi_scope=asgi_scope, root_path_in_path=root_path_in_path),
        )

    client = asgi_scope.get("client")
    if client and should_send_default_pii():
        request_data["env"] = {"REMOTE_ADDR": _get_ip(asgi_scope)}

    return request_data


def _get_request_attributes(
    asgi_scope: "Any",
    root_path_in_path: "_RootPathInPath",
) -> "dict[str, Any]":
    """
    Return attributes related to the HTTP request from the ASGI scope.
    """
    attributes: "dict[str, Any]" = {}

    ty = asgi_scope["type"]
    if ty in ("http", "websocket"):
        if asgi_scope.get("method"):
            attributes["http.request.method"] = asgi_scope["method"].upper()

        headers = _filter_headers(_get_headers(asgi_scope), use_annotated_value=False)
        for header, value in headers.items():
            attributes[f"http.request.header.{header.lower()}"] = value

        if should_send_default_pii():
            query = _get_query(asgi_scope)
            if query:
                attributes["http.query"] = query

            path = _get_path(asgi_scope=asgi_scope, root_path_in_path=root_path_in_path)
            attributes["url.path"] = path

            url_without_query_string = _get_url(
                asgi_scope,
                "http" if ty == "http" else "ws",
                headers.get("host"),
                path=path,
            )
            query_string = _get_query(asgi_scope)
            attributes["url.full"] = (
                f"{url_without_query_string}?{query_string}"
                if query_string is not None
                else url_without_query_string
            )

    client = asgi_scope.get("client")
    if client and should_send_default_pii():
        ip = _get_ip(asgi_scope)
        attributes["client.address"] = ip

    return attributes



from typing import Any, AsyncIterable, Callable, Union

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.integrations.grpc.consts import SPAN_ORIGIN
from sentry_sdk.tracing_utils import has_span_streaming_enabled

try:
    from google.protobuf.message import Message
    from grpc.aio import (
        ClientCallDetails,
        Metadata,
        UnaryStreamCall,
        UnaryStreamClientInterceptor,
        UnaryUnaryCall,
        UnaryUnaryClientInterceptor,
    )
except ImportError:
    raise DidNotEnable("grpcio is not installed")


class ClientInterceptor:
    @staticmethod
    def _update_client_call_details_metadata_from_scope(
        client_call_details: "ClientCallDetails",
    ) -> "ClientCallDetails":
        if client_call_details.metadata is None:
            client_call_details = client_call_details._replace(metadata=Metadata())
        elif not isinstance(client_call_details.metadata, Metadata):
            # This is a workaround for a GRPC bug, which was fixed in grpcio v1.60.0
            # See https://github.com/grpc/grpc/issues/34298.
            client_call_details = client_call_details._replace(
                metadata=Metadata.from_tuple(client_call_details.metadata)
            )
        for (
            key,
            value,
        ) in sentry_sdk.get_current_scope().iter_trace_propagation_headers():
            client_call_details.metadata.add(key, value)
        return client_call_details


class SentryUnaryUnaryClientInterceptor(ClientInterceptor, UnaryUnaryClientInterceptor):  # type: ignore
    async def intercept_unary_unary(
        self,
        continuation: "Callable[[ClientCallDetails, Message], UnaryUnaryCall]",
        client_call_details: "ClientCallDetails",
        request: "Message",
    ) -> "Union[UnaryUnaryCall, Message]":
        method = client_call_details.method

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )
                return await continuation(client_call_details, request)
            with sentry_sdk.traces.start_span(
                name="unary unary call to %s" % method.decode(),
                attributes={
                    "sentry.op": OP.GRPC_CLIENT,
                    "sentry.origin": SPAN_ORIGIN,
                    SPANDATA.RPC_METHOD: method.decode(),
                },
            ) as span:
                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )

                response = await continuation(client_call_details, request)
                status_code = await response.code()
                span.set_attribute(SPANDATA.RPC_RESPONSE_STATUS_CODE, status_code.name)

                return response
        else:
            with sentry_sdk.start_span(
                op=OP.GRPC_CLIENT,
                name="unary unary call to %s" % method.decode(),
                origin=SPAN_ORIGIN,
            ) as span:
                span.set_data("type", "unary unary")
                span.set_data("method", method)

                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )

                response = await continuation(client_call_details, request)
                status_code = await response.code()
                span.set_data("code", status_code.name)

                return response


class SentryUnaryStreamClientInterceptor(
    ClientInterceptor,
    UnaryStreamClientInterceptor,  # type: ignore
):
    async def intercept_unary_stream(
        self,
        continuation: "Callable[[ClientCallDetails, Message], UnaryStreamCall]",
        client_call_details: "ClientCallDetails",
        request: "Message",
    ) -> "Union[AsyncIterable[Any], UnaryStreamCall]":
        method = client_call_details.method

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )
                return await continuation(client_call_details, request)
            with sentry_sdk.traces.start_span(
                name="unary stream call to %s" % method.decode(),
                attributes={
                    "sentry.op": OP.GRPC_CLIENT,
                    "sentry.origin": SPAN_ORIGIN,
                    SPANDATA.RPC_METHOD: method.decode(),
                },
            ) as span:
                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )

                response = await continuation(client_call_details, request)

                return response
        else:
            with sentry_sdk.start_span(
                op=OP.GRPC_CLIENT,
                name="unary stream call to %s" % method.decode(),
                origin=SPAN_ORIGIN,
            ) as span:
                span.set_data("type", "unary stream")
                span.set_data("method", method)

                client_call_details = (
                    self._update_client_call_details_metadata_from_scope(
                        client_call_details
                    )
                )

                response = await continuation(client_call_details, request)
                # status_code = await response.code()
                # span.set_data("code", status_code)

                return response



import functools
import sys

import sentry_sdk
from sentry_sdk._types import TYPE_CHECKING
from sentry_sdk.integrations import Integration
from sentry_sdk.utils import capture_internal_exceptions, event_from_exception

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import NoReturn, Union


class SysExitIntegration(Integration):
    """Captures sys.exit calls and sends them as events to Sentry.

    By default, SystemExit exceptions are not captured by the SDK. Enabling this integration will capture SystemExit
    exceptions generated by sys.exit calls and send them to Sentry.

    This integration, in its default configuration, only captures the sys.exit call if the exit code is a non-zero and
    non-None value (unsuccessful exits). Pass `capture_successful_exits=True` to capture successful exits as well.
    Note that the integration does not capture SystemExit exceptions raised outside a call to sys.exit.
    """

    identifier = "sys_exit"

    def __init__(self, *, capture_successful_exits: bool = False) -> None:
        self._capture_successful_exits = capture_successful_exits

    @staticmethod
    def setup_once() -> None:
        SysExitIntegration._patch_sys_exit()

    @staticmethod
    def _patch_sys_exit() -> None:
        old_exit: "Callable[[Union[str, int, None]], NoReturn]" = sys.exit

        @functools.wraps(old_exit)
        def sentry_patched_exit(__status: "Union[str, int, None]" = 0) -> "NoReturn":
            # @ensure_integration_enabled ensures that this is non-None
            integration = sentry_sdk.get_client().get_integration(SysExitIntegration)
            if integration is None:
                old_exit(__status)

            try:
                old_exit(__status)
            except SystemExit as e:
                with capture_internal_exceptions():
                    if integration._capture_successful_exits or __status not in (
                        0,
                        None,
                    ):
                        _capture_exception(e)
                raise e

        sys.exit = sentry_patched_exit


def _capture_exception(exc: "SystemExit") -> None:
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": SysExitIntegration.identifier, "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)



from sentry_sdk.integrations import DidNotEnable

try:
    from opentelemetry.context import create_key
except ImportError:
    raise DidNotEnable("opentelemetry not installed")

SENTRY_TRACE_KEY = create_key("sentry-trace")
SENTRY_BAGGAGE_KEY = create_key("sentry-baggage")



import socket

import sentry_sdk
from sentry_sdk._types import MYPY
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import Integration
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if MYPY:
    from socket import AddressFamily, SocketKind
    from typing import List, Optional, Tuple, Union

__all__ = ["SocketIntegration"]


class SocketIntegration(Integration):
    identifier = "socket"
    origin = f"auto.socket.{identifier}"

    @staticmethod
    def setup_once() -> None:
        """
        patches two of the most used functions of socket: create_connection and getaddrinfo(dns resolver)
        """
        _patch_create_connection()
        _patch_getaddrinfo()


def _get_span_description(
    host: "Union[bytes, str, None]", port: "Union[bytes, str, int, None]"
) -> str:
    try:
        host = host.decode()  # type: ignore
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        port = port.decode()  # type: ignore
    except (UnicodeDecodeError, AttributeError):
        pass

    description = "%s:%s" % (host, port)  # type: ignore
    return description


def _patch_create_connection() -> None:
    real_create_connection = socket.create_connection

    def create_connection(
        address: "Tuple[Optional[str], int]",
        timeout: "Optional[float]" = socket._GLOBAL_DEFAULT_TIMEOUT,  # type: ignore
        source_address: "Optional[Tuple[Union[bytearray, bytes, str], int]]" = None,
    ) -> "socket.socket":
        client = sentry_sdk.get_client()
        integration = client.get_integration(SocketIntegration)
        if integration is None:
            return real_create_connection(address, timeout, source_address)

        if has_span_streaming_enabled(client.options):
            if sentry_sdk.traces.get_current_span() is None:
                return real_create_connection(address, timeout, source_address)

            with sentry_sdk.traces.start_span(
                name=_get_span_description(address[0], address[1]),
                attributes={
                    "sentry.op": OP.SOCKET_CONNECTION,
                    "sentry.origin": SocketIntegration.origin,
                },
            ) as span:
                if address[0] is not None:
                    span.set_attribute(SPANDATA.SERVER_ADDRESS, address[0])
                span.set_attribute(SPANDATA.SERVER_PORT, address[1])

                return real_create_connection(
                    address=address, timeout=timeout, source_address=source_address
                )
        else:
            with sentry_sdk.start_span(
                op=OP.SOCKET_CONNECTION,
                name=_get_span_description(address[0], address[1]),
                origin=SocketIntegration.origin,
            ) as span:
                span.set_data("address", address)
                span.set_data("timeout", timeout)
                span.set_data("source_address", source_address)

                return real_create_connection(
                    address=address, timeout=timeout, source_address=source_address
                )

    socket.create_connection = create_connection  # type: ignore


def _patch_getaddrinfo() -> None:
    real_getaddrinfo = socket.getaddrinfo

    def getaddrinfo(
        host: "Union[bytes, str, None]",
        port: "Union[bytes, str, int, None]",
        family: int = 0,
        type: int = 0,
        proto: int = 0,
        flags: int = 0,
    ) -> "List[Tuple[AddressFamily, SocketKind, int, str, Union[Tuple[str, int], Tuple[str, int, int, int], Tuple[int, bytes]]]]":
        client = sentry_sdk.get_client()
        integration = client.get_integration(SocketIntegration)
        if integration is None:
            return real_getaddrinfo(host, port, family, type, proto, flags)

        if has_span_streaming_enabled(client.options):
            if sentry_sdk.traces.get_current_span() is None:
                return real_getaddrinfo(host, port, family, type, proto, flags)

            with sentry_sdk.traces.start_span(
                name=_get_span_description(host, port),
                attributes={
                    "sentry.op": OP.SOCKET_DNS,
                    "sentry.origin": SocketIntegration.origin,
                },
            ) as span:
                if isinstance(host, str):
                    span.set_attribute(SPANDATA.SERVER_ADDRESS, host)
                elif isinstance(host, bytes):
                    span.set_attribute(
                        SPANDATA.SERVER_ADDRESS, host.decode(errors="replace")
                    )

                if isinstance(port, int):
                    span.set_attribute(SPANDATA.SERVER_PORT, port)
                elif port is not None:
                    try:
                        span.set_attribute(SPANDATA.SERVER_PORT, int(port))
                    except (ValueError, TypeError):
                        pass

                return real_getaddrinfo(host, port, family, type, proto, flags)
        else:
            with sentry_sdk.start_span(
                op=OP.SOCKET_DNS,
                name=_get_span_description(host, port),
                origin=SocketIntegration.origin,
            ) as span:
                span.set_data("host", host)
                span.set_data("port", port)

                return real_getaddrinfo(host, port, family, type, proto, flags)

    socket.getaddrinfo = getaddrinfo



import atexit
import os
import random
import sys
import threading
import time
import uuid
import warnings
from collections import deque
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from sentry_sdk._lru_cache import LRUCache
from sentry_sdk.consts import VERSION
from sentry_sdk.envelope import Envelope
from sentry_sdk.profiler.utils import (
    DEFAULT_SAMPLING_FREQUENCY,
    extract_stack,
)
from sentry_sdk.utils import (
    capture_internal_exception,
    is_gevent,
    logger,
    now,
    set_in_app_in_frames,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Deque, Dict, List, Optional, Set, Type, Union

    from typing_extensions import TypedDict

    from sentry_sdk._types import ContinuousProfilerMode, SDKInfo
    from sentry_sdk.profiler.utils import (
        ExtractedSample,
        FrameId,
        ProcessedFrame,
        ProcessedStack,
        StackId,
        ThreadId,
    )

    ProcessedSample = TypedDict(
        "ProcessedSample",
        {
            "timestamp": float,
            "thread_id": ThreadId,
            "stack_id": int,
        },
    )


try:
    from gevent.monkey import get_original
    from gevent.threadpool import ThreadPool as _ThreadPool

    ThreadPool: "Optional[Type[_ThreadPool]]" = _ThreadPool
    thread_sleep = get_original("time", "sleep")
except ImportError:
    thread_sleep = time.sleep
    ThreadPool = None


_scheduler: "Optional[ContinuousScheduler]" = None


def setup_continuous_profiler(
    options: "Dict[str, Any]",
    sdk_info: "SDKInfo",
    capture_func: "Callable[[Envelope], None]",
) -> bool:
    global _scheduler

    already_initialized = _scheduler is not None

    if already_initialized:
        logger.debug("[Profiling] Continuous Profiler is already setup")
        teardown_continuous_profiler()

    if is_gevent():
        # If gevent has patched the threading modules then we cannot rely on
        # them to spawn a native thread for sampling.
        # Instead we default to the GeventContinuousScheduler which is capable of
        # spawning native threads within gevent.
        default_profiler_mode = GeventContinuousScheduler.mode
    else:
        default_profiler_mode = ThreadContinuousScheduler.mode

    if options.get("profiler_mode") is not None:
        profiler_mode = options["profiler_mode"]
    else:
        # TODO: deprecate this and just use the existing `profiler_mode`
        experiments = options.get("_experiments", {})

        profiler_mode = (
            experiments.get("continuous_profiling_mode") or default_profiler_mode
        )

    frequency = DEFAULT_SAMPLING_FREQUENCY

    if profiler_mode == ThreadContinuousScheduler.mode:
        _scheduler = ThreadContinuousScheduler(
            frequency, options, sdk_info, capture_func
        )
    elif profiler_mode == GeventContinuousScheduler.mode:
        _scheduler = GeventContinuousScheduler(
            frequency, options, sdk_info, capture_func
        )
    else:
        raise ValueError("Unknown continuous profiler mode: {}".format(profiler_mode))

    logger.debug(
        "[Profiling] Setting up continuous profiler in {mode} mode".format(
            mode=_scheduler.mode
        )
    )

    if not already_initialized:
        atexit.register(teardown_continuous_profiler)

    return True


def is_profile_session_sampled() -> bool:
    if _scheduler is None:
        return False
    return _scheduler.sampled


def try_autostart_continuous_profiler() -> None:
    # TODO: deprecate this as it'll be replaced by the auto lifecycle option

    if _scheduler is None:
        return

    if not _scheduler.is_auto_start_enabled():
        return

    _scheduler.manual_start()


def try_profile_lifecycle_trace_start() -> "Union[ContinuousProfile, None]":
    if _scheduler is None:
        return None

    return _scheduler.auto_start()


def start_profiler() -> None:
    if _scheduler is None:
        return

    _scheduler.manual_start()


def start_profile_session() -> None:
    warnings.warn(
        "The `start_profile_session` function is deprecated. Please use `start_profile` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    start_profiler()


def stop_profiler() -> None:
    if _scheduler is None:
        return

    _scheduler.manual_stop()


def stop_profile_session() -> None:
    warnings.warn(
        "The `stop_profile_session` function is deprecated. Please use `stop_profile` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    stop_profiler()


def teardown_continuous_profiler() -> None:
    stop_profiler()

    global _scheduler
    _scheduler = None


def get_profiler_id() -> "Union[str, None]":
    if _scheduler is None:
        return None
    return _scheduler.profiler_id


def determine_profile_session_sampling_decision(
    sample_rate: "Union[float, None]",
) -> bool:
    # `None` is treated as `0.0`
    if not sample_rate:
        return False

    return random.random() < float(sample_rate)


class ContinuousProfile:
    active: bool = True

    def stop(self) -> None:
        self.active = False


class ContinuousScheduler:
    mode: "ContinuousProfilerMode" = "unknown"

    def __init__(
        self,
        frequency: int,
        options: "Dict[str, Any]",
        sdk_info: "SDKInfo",
        capture_func: "Callable[[Envelope], None]",
    ) -> None:
        self.interval = 1.0 / frequency
        self.options = options
        self.sdk_info = sdk_info
        self.capture_func = capture_func

        self.lifecycle = self.options.get("profile_lifecycle")
        profile_session_sample_rate = self.options.get("profile_session_sample_rate")
        self.sampled = determine_profile_session_sampling_decision(
            profile_session_sample_rate
        )

        self.sampler = self.make_sampler()
        self.buffer: "Optional[ProfileBuffer]" = None
        self.pid: "Optional[int]" = None

        self.running = False
        self.soft_shutdown = False

        self.new_profiles: "Deque[ContinuousProfile]" = deque(maxlen=128)
        self.active_profiles: "Set[ContinuousProfile]" = set()

    def is_auto_start_enabled(self) -> bool:
        # Ensure that the scheduler only autostarts once per process.
        # This is necessary because many web servers use forks to spawn
        # additional processes. And the profiler is only spawned on the
        # master process, then it often only profiles the main process
        # and not the ones where the requests are being handled.
        if self.pid == os.getpid():
            return False

        experiments = self.options.get("_experiments")
        if not experiments:
            return False

        return experiments.get("continuous_profiling_auto_start")

    def auto_start(self) -> "Union[ContinuousProfile, None]":
        if not self.sampled:
            return None

        if self.lifecycle != "trace":
            return None

        logger.debug("[Profiling] Auto starting profiler")

        profile = ContinuousProfile()

        self.new_profiles.append(profile)
        self.ensure_running()

        return profile

    def manual_start(self) -> None:
        if not self.sampled:
            return

        if self.lifecycle != "manual":
            return

        self.ensure_running()

    def manual_stop(self) -> None:
        if self.lifecycle != "manual":
            return

        self.teardown()

    def ensure_running(self) -> None:
        raise NotImplementedError

    def teardown(self) -> None:
        raise NotImplementedError

    def pause(self) -> None:
        raise NotImplementedError

    def reset_buffer(self) -> None:
        self.buffer = ProfileBuffer(
            self.options, self.sdk_info, PROFILE_BUFFER_SECONDS, self.capture_func
        )

    @property
    def profiler_id(self) -> "Union[str, None]":
        if not self.running or self.buffer is None:
            return None
        return self.buffer.profiler_id

    def make_sampler(self) -> "Callable[..., bool]":
        cwd = os.getcwd()

        cache = LRUCache(max_size=256)

        if self.lifecycle == "trace":

            def _sample_stack(*args: "Any", **kwargs: "Any") -> bool:
                """
                Take a sample of the stack on all the threads in the process.
                This should be called at a regular interval to collect samples.
                """

                # no profiles taking place, so we can stop early
                if not self.new_profiles and not self.active_profiles:
                    return True

                # This is the number of profiles we want to pop off.
                # It's possible another thread adds a new profile to
                # the list and we spend longer than we want inside
                # the loop below.
                #
                # Also make sure to set this value before extracting
                # frames so we do not write to any new profiles that
                # were started after this point.
                new_profiles = len(self.new_profiles)

                ts = now()

                try:
                    sample = [
                        (str(tid), extract_stack(frame, cache, cwd))
                        for tid, frame in sys._current_frames().items()
                    ]
                except AttributeError:
                    # For some reason, the frame we get doesn't have certain attributes.
                    # When this happens, we abandon the current sample as it's bad.
                    capture_internal_exception(sys.exc_info())
                    return False

                # Move the new profiles into the active_profiles set.
                #
                # We cannot directly add the to active_profiles set
                # in `start_profiling` because it is called from other
                # threads which can cause a RuntimeError when it the
                # set sizes changes during iteration without a lock.
                #
                # We also want to avoid using a lock here so threads
                # that are starting profiles are not blocked until it
                # can acquire the lock.
                for _ in range(new_profiles):
                    self.active_profiles.add(self.new_profiles.popleft())
                inactive_profiles = []

                for profile in self.active_profiles:
                    if not profile.active:
                        # If a profile is marked inactive, we buffer it
                        # to `inactive_profiles` so it can be removed.
                        # We cannot remove it here as it would result
                        # in a RuntimeError.
                        inactive_profiles.append(profile)

                for profile in inactive_profiles:
                    self.active_profiles.remove(profile)

                if self.buffer is not None:
                    self.buffer.write(ts, sample)

                return False

        else:

            def _sample_stack(*args: "Any", **kwargs: "Any") -> bool:
                """
                Take a sample of the stack on all the threads in the process.
                This should be called at a regular interval to collect samples.
                """

                ts = now()

                try:
                    sample = [
                        (str(tid), extract_stack(frame, cache, cwd))
                        for tid, frame in sys._current_frames().items()
                    ]
                except AttributeError:
                    # For some reason, the frame we get doesn't have certain attributes.
                    # When this happens, we abandon the current sample as it's bad.
                    capture_internal_exception(sys.exc_info())
                    return False

                if self.buffer is not None:
                    self.buffer.write(ts, sample)

                return False

        return _sample_stack

    def run(self) -> None:
        last = time.perf_counter()

        while self.running:
            self.soft_shutdown = self.sampler()

            # some time may have elapsed since the last time
            # we sampled, so we need to account for that and
            # not sleep for too long
            elapsed = time.perf_counter() - last
            if elapsed < self.interval:
                thread_sleep(self.interval - elapsed)

            # the soft shutdown happens here to give it a chance
            # for the profiler to be reused
            if self.soft_shutdown:
                self.running = False

                # make sure to explicitly exit the profiler here or there might
                # be multiple profilers at once
                break

            # after sleeping, make sure to take the current
            # timestamp so we can use it next iteration
            last = time.perf_counter()

        buffer = self.buffer
        if buffer is not None:
            buffer.flush()


class ThreadContinuousScheduler(ContinuousScheduler):
    """
    This scheduler is based on running a daemon thread that will call
    the sampler at a regular interval.
    """

    mode: "ContinuousProfilerMode" = "thread"
    name = "sentry.profiler.ThreadContinuousScheduler"

    def __init__(
        self,
        frequency: int,
        options: "Dict[str, Any]",
        sdk_info: "SDKInfo",
        capture_func: "Callable[[Envelope], None]",
    ) -> None:
        super().__init__(frequency, options, sdk_info, capture_func)

        self.thread: "Optional[threading.Thread]" = None
        self.lock = threading.Lock()

    def ensure_running(self) -> None:
        self.soft_shutdown = False

        pid = os.getpid()

        # is running on the right process
        if self.running and self.pid == pid:
            return

        with self.lock:
            # another thread may have tried to acquire the lock
            # at the same time so it may start another thread
            # make sure to check again before proceeding
            if self.running and self.pid == pid:
                return

            self.pid = pid
            self.running = True

            # if the profiler thread is changing,
            # we should create a new buffer along with it
            self.reset_buffer()

            # make sure the thread is a daemon here otherwise this
            # can keep the application running after other threads
            # have exited
            self.thread = threading.Thread(name=self.name, target=self.run, daemon=True)

            try:
                self.thread.start()
            except RuntimeError:
                # Unfortunately at this point the interpreter is in a state that no
                # longer allows us to spawn a thread and we have to bail.
                self.running = False
                self.thread = None

    def teardown(self) -> None:
        if self.running:
            self.running = False

        if self.thread is not None:
            self.thread.join()
            self.thread = None


class GeventContinuousScheduler(ContinuousScheduler):
    """
    This scheduler is based on the thread scheduler but adapted to work with
    gevent. When using gevent, it may monkey patch the threading modules
    (`threading` and `_thread`). This results in the use of greenlets instead
    of native threads.

    This is an issue because the sampler CANNOT run in a greenlet because
    1. Other greenlets doing sync work will prevent the sampler from running
    2. The greenlet runs in the same thread as other greenlets so when taking
       a sample, other greenlets will have been evicted from the thread. This
       results in a sample containing only the sampler's code.
    """

    mode: "ContinuousProfilerMode" = "gevent"

    def __init__(
        self,
        frequency: int,
        options: "Dict[str, Any]",
        sdk_info: "SDKInfo",
        capture_func: "Callable[[Envelope], None]",
    ) -> None:
        if ThreadPool is None:
            raise ValueError("Profiler mode: {} is not available".format(self.mode))

        super().__init__(frequency, options, sdk_info, capture_func)

        self.thread: "Optional[_ThreadPool]" = None
        self.lock = threading.Lock()

    def ensure_running(self) -> None:
        self.soft_shutdown = False

        pid = os.getpid()

        # is running on the right process
        if self.running and self.pid == pid:
            return

        with self.lock:
            # another thread may have tried to acquire the lock
            # at the same time so it may start another thread
            # make sure to check again before proceeding
            if self.running and self.pid == pid:
                return

            self.pid = pid
            self.running = True

            # if the profiler thread is changing,
            # we should create a new buffer along with it
            self.reset_buffer()

            self.thread = ThreadPool(1)  # type: ignore[misc]
            try:
                self.thread.spawn(self.run)
            except RuntimeError:
                # Unfortunately at this point the interpreter is in a state that no
                # longer allows us to spawn a thread and we have to bail.
                self.running = False
                self.thread = None

    def teardown(self) -> None:
        if self.running:
            self.running = False

        if self.thread is not None:
            self.thread.join()
            self.thread = None


PROFILE_BUFFER_SECONDS = 60


class ProfileBuffer:
    def __init__(
        self,
        options: "Dict[str, Any]",
        sdk_info: "SDKInfo",
        buffer_size: int,
        capture_func: "Callable[[Envelope], None]",
    ) -> None:
        self.options = options
        self.sdk_info = sdk_info
        self.buffer_size = buffer_size
        self.capture_func = capture_func

        self.profiler_id = uuid.uuid4().hex
        self.chunk = ProfileChunk()

        # Make sure to use the same clock to compute a sample's monotonic timestamp
        # to ensure the timestamps are correctly aligned.
        self.start_monotonic_time = now()

        # Make sure the start timestamp is defined only once per profiler id.
        # This prevents issues with clock drift within a single profiler session.
        #
        # Subtracting the start_monotonic_time here to find a fixed starting position
        # for relative monotonic timestamps for each sample.
        self.start_timestamp = (
            datetime.now(timezone.utc).timestamp() - self.start_monotonic_time
        )

    def write(self, monotonic_time: float, sample: "ExtractedSample") -> None:
        if self.should_flush(monotonic_time):
            self.flush()
            self.chunk = ProfileChunk()
            self.start_monotonic_time = now()

        self.chunk.write(self.start_timestamp + monotonic_time, sample)

    def should_flush(self, monotonic_time: float) -> bool:
        # If the delta between the new monotonic time and the start monotonic time
        # exceeds the buffer size, it means we should flush the chunk
        return monotonic_time - self.start_monotonic_time >= self.buffer_size

    def flush(self) -> None:
        chunk = self.chunk.to_json(self.profiler_id, self.options, self.sdk_info)
        envelope = Envelope()
        envelope.add_profile_chunk(chunk)
        self.capture_func(envelope)


class ProfileChunk:
    def __init__(self) -> None:
        self.chunk_id = uuid.uuid4().hex

        self.indexed_frames: "Dict[FrameId, int]" = {}
        self.indexed_stacks: "Dict[StackId, int]" = {}
        self.frames: "List[ProcessedFrame]" = []
        self.stacks: "List[ProcessedStack]" = []
        self.samples: "List[ProcessedSample]" = []

    def write(self, ts: float, sample: "ExtractedSample") -> None:
        for tid, (stack_id, frame_ids, frames) in sample:
            try:
                # Check if the stack is indexed first, this lets us skip
                # indexing frames if it's not necessary
                if stack_id not in self.indexed_stacks:
                    for i, frame_id in enumerate(frame_ids):
                        if frame_id not in self.indexed_frames:
                            self.indexed_frames[frame_id] = len(self.indexed_frames)
                            self.frames.append(frames[i])

                    self.indexed_stacks[stack_id] = len(self.indexed_stacks)
                    self.stacks.append(
                        [self.indexed_frames[frame_id] for frame_id in frame_ids]
                    )

                self.samples.append(
                    {
                        "timestamp": ts,
                        "thread_id": tid,
                        "stack_id": self.indexed_stacks[stack_id],
                    }
                )
            except AttributeError:
                # For some reason, the frame we get doesn't have certain attributes.
                # When this happens, we abandon the current sample as it's bad.
                capture_internal_exception(sys.exc_info())

    def to_json(
        self, profiler_id: str, options: "Dict[str, Any]", sdk_info: "SDKInfo"
    ) -> "Dict[str, Any]":
        profile = {
            "frames": self.frames,
            "stacks": self.stacks,
            "samples": self.samples,
            "thread_metadata": {
                str(thread.ident): {
                    "name": str(thread.name),
                }
                for thread in threading.enumerate()
            },
        }

        set_in_app_in_frames(
            profile["frames"],
            options["in_app_exclude"],
            options["in_app_include"],
            options["project_root"],
        )

        payload = {
            "chunk_id": self.chunk_id,
            "client_sdk": {
                "name": sdk_info["name"],
                "version": VERSION,
            },
            "platform": "python",
            "profile": profile,
            "profiler_id": profiler_id,
            "version": "2",
        }

        for key in "release", "environment", "dist":
            if options[key] is not None:
                payload[key] = str(options[key]).strip()

        return payload



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, TypeVar

    T = TypeVar("T")


PY37 = sys.version_info[0] == 3 and sys.version_info[1] >= 7
PY38 = sys.version_info[0] == 3 and sys.version_info[1] >= 8
PY310 = sys.version_info[0] == 3 and sys.version_info[1] >= 10
PY311 = sys.version_info[0] == 3 and sys.version_info[1] >= 11


def with_metaclass(meta: "Any", *bases: "Any") -> "Any":
    class MetaClass(type):
        def __new__(metacls: "Any", name: "Any", this_bases: "Any", d: "Any") -> "Any":
            return meta(name, bases, d)

    return type.__new__(MetaClass, "temporary_class", (), {})


def check_uwsgi_thread_support() -> bool:
    # We check two things here:
    #
    # 1. uWSGI doesn't run in threaded mode by default -- issue a warning if
    #    that's the case.
    #
    # 2. Additionally, if uWSGI is running in preforking mode (default), it needs
    #    the --py-call-uwsgi-fork-hooks option for the SDK to work properly. This
    #    is because any background threads spawned before the main process is
    #    forked are NOT CLEANED UP IN THE CHILDREN BY DEFAULT even if
    #    --enable-threads is on. One has to explicitly provide
    #    --py-call-uwsgi-fork-hooks to force uWSGI to run regular cpython
    #    after-fork hooks that take care of cleaning up stale thread data.
    try:
        from uwsgi import opt  # type: ignore
    except ImportError:
        return True

    from sentry_sdk.consts import FALSE_VALUES

    def enabled(option: str) -> bool:
        value = opt.get(option, False)
        if isinstance(value, bool):
            return value

        if isinstance(value, bytes):
            try:
                value = value.decode()
            except Exception:
                pass

        return value and str(value).lower() not in FALSE_VALUES  # type: ignore[return-value]

    # When `threads` is passed in as a uwsgi option,
    # `enable-threads` is implied on.
    threads_enabled = "threads" in opt or enabled("enable-threads")
    fork_hooks_on = enabled("py-call-uwsgi-fork-hooks")
    lazy_mode = enabled("lazy-apps") or enabled("lazy")

    if lazy_mode and not threads_enabled:
        from warnings import warn

        warn(
            Warning(
                "IMPORTANT: "
                "We detected the use of uWSGI without thread support. "
                "This might lead to unexpected issues. "
                'Please run uWSGI with "--enable-threads" for full support.'
            )
        )

        return False

    elif not lazy_mode and (not threads_enabled or not fork_hooks_on):
        from warnings import warn

        warn(
            Warning(
                "IMPORTANT: "
                "We detected the use of uWSGI in preforking mode without "
                "thread support. This might lead to crashing workers. "
                'Please run uWSGI with both "--enable-threads" and '
                '"--py-call-uwsgi-fork-hooks" for full support.'
            )
        )

        return False

    return True
